//      Copyright (c) ClearSpeed Technology plc 2000,2002,2006
//      All rights reserved.
//
//   This file is part of the ClearSpeed Open Source Driver
//
//   The ClearSpeed Open Source Driver is free software; you can redistribute it
//   and/or modify it under the terms of the GNU General Public License as
//   published by the Free Software Foundation; either version 2 of the License,
//   or (at your option) any later version.
//
//   The ClearSpeed Open Source Driver is distributed in the hope that it will
//   be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
//   Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with The ClearSpeed Open Source Driver; if not, write to the Free Software
//   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2006 Silicon Graphics, Inc.  All Rights Reserved.
 */

#include <linux/module.h>
#include <linux/ctype.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <asm/page.h>

/* Description of user side interface, shared with user space. */
#include "csx_driver.h"

#define CSX_DEBUG 0
#define CSX_DEBUG_IOCTLS 0
#define CSX_DEBUG_EXTREME 0

/* set to 1 to get a coherent data bounce buffer for DMA */
#define CSX_BOUNCE 0
/* set to 1 to get kernel mapping for memory aperture and read/write syscall to it */
/*
   NB you cannot afford the kernel vmalloc space with more than 1 card
   and the read/write syscalls are not required
 */
#define CS_NOMAPMEM 0

#if defined (CONFIG_ARCH_IA64_SN2) || defined (CONFIG_ARCH_IA64_GENERIC)
#define MMIOWB() do { mmiowb(); } while (0)
#else
#define MMIOWB()
#endif

/******************************
 * Definitions and data types *
 ******************************/

#define CSX_CTL_REGSIZE sizeof(uint32_t)

#define CSX_CTL_DEVNUM	0
#define CSX_MEM_DEVNUM	1
#define CSX_NUM_DEVS	2	/* Number of char special devices per board */

/* Indexes into control aperture of various interesting registers.
 * Assumes 32-bit registers.
 */
#define CSX_CTL_PCIINTR_INDX		(0x0>>2)	/* ctl + 0x0 */
#define CSX_CTL_DMAINTR_INDX		(0x4>>2)	/* ctl + 0x4 */
#define CSX_CTL_PCIINTRMASK_INDX	(0x10>>2)	/* ctl + 0x10 */
#define CSX_CTL_DMAINTRMASK_INDX	(0x14>>2)	/* ctl + 0x14 */
#define CSX_DMA_BYTECOUNT_INDX      (0x1000>>2)	/* ctl + 0x1000 */
#define CSX_DMA_CONTROL_INDX        (0x1004>>2)	/* ctl + 0x1004 */
#define CSX_DMA_FPGA_ADDR           (0x1008>>2)	/* ctl + 0x1008 */
#define CSX_DMA_PCI_ADDR            (0x1010>>2)	/* ctl + 0x1010 */
#define CSX_DMA_DESC_INDX           (0x1018>>2)	/* ctl + 0x1018 Descriptor chain address */
#define CSX_CTL_FPGA_VERSION_INDX	(0x80000>>2)	/* ctl + 0x80000 */

#define CSX_CTL_PCIINTR_MASK	0x1	/* MTAP generated interrupts */
#define CSX_CTL_DMAINTR_MASK	0x7	/* DMA engine interrupts */

/* initially copied from st.c. */
struct csx_dma_region {
	unsigned int n_pages[2];	                /* # of kernel pages */
	struct scatterlist *sglist[2];	        /* IOMMU mapping */
	unsigned int pages_pinned[2];	        /* Actual # to pci_map_sg (nents) */
	enum dma_data_direction direction;	/* Direction of DMA transfer */

    /* Information on the coherent data mapping if used. */
	dma_addr_t data_bus_address;	        /* pci bus address for device. */
	void *data_kaddr;	                /* kernel map address for buffer. */
	unsigned data_buf_size;	                /* Size of the coherent buffer. */

    /* Information on the coherent descriptor mapping */
	dma_addr_t desc_bus_address[2];	        /* pci bus address for device. */
	void *desc_kaddr;	                /* kernel map address for descriptor buffer. */
	unsigned desc_buf_size;	                /* Size of the coherent descriptor buffer. */
};

/* All the driver-tracked information about this device instance */
struct csx_device_data;

struct csx_ctl_data {
	struct csx_device_data *maindata;
	dev_t dev;
	struct cdev cdev;
	struct class_device classdev;
	atomic_t intr_count;
	wait_queue_head_t intr_queue;
	unsigned long bar;	/* PCI address map */
	unsigned long len;
	uint32_t *regs;	/* from ioremap_nocache */
};

struct csx_mem_data {
	struct csx_device_data *maindata;
	dev_t dev;
	struct cdev cdev;
	struct class_device classdev;
	atomic_t intr_count;
	wait_queue_head_t intr_queue;
	unsigned long bar;	/* PCI address map */
	unsigned long len;
	unsigned char *mem;	/* from ioremap_nocache */
	struct csx_dma_region region;
};

struct csx_device_data {
	struct pci_dev *pcidev;
	dev_t dev_base;
	struct csx_ctl_data *ctl;
	struct csx_mem_data *mem;
};

/* Note: to get to the generic dev handle do: &xxx->pcidev.dev */

/* Per-file data for control register device */
struct csx_ctl_file {
	__s32 intr_count;
};

/* Per-file data for memory device */
struct csx_mem_file {
	__s32 intr_count;
};

/***************
 * Module data *
 ***************/

static dev_t firstdev;		/* First dynamic device number */
static dev_t nextdev;		/* Next available dynamic device number */
static spinlock_t nextdev_lock;

/******************
 * Class routines *
 ******************/

/* Frees per-instance device data when all references to it go away. */
static void
csx_ctl_class_release(struct class_device *classdev)
{
	struct csx_ctl_data *ccd;

	ccd = container_of(classdev, struct csx_ctl_data, classdev);

	cdev_del(&ccd->cdev);
	iounmap(ccd->regs);
	release_mem_region(ccd->bar, ccd->len);
	kfree(ccd);
}

static struct class csx_ctl_class = {
	.name = "csxctl",
	.release = csx_ctl_class_release,
};

/* Frees per-instance device data when all references to it go away. */
static void
csx_mem_class_release(struct class_device *classdev)
{
	struct csx_mem_data *cmd;
	struct csx_device_data *cdd;

	cmd = container_of(classdev, struct csx_mem_data, classdev);

	cdd = cmd->maindata;

	cdev_del(&cmd->cdev);
#if CS_NOMAPMEM
	iounmap(cmd->mem);
#endif
	release_mem_region(cmd->bar, cmd->len);
#if CSX_BOUNCE
	dma_free_coherent(&cdd->pcidev->dev, cdd->mem->region.data_buf_size,
			  cdd->mem->region.data_kaddr,
			  cdd->mem->region.data_bus_address);
#endif
	dma_free_coherent(&cdd->pcidev->dev, cdd->mem->region.desc_buf_size * 2,
			  cdd->mem->region.desc_kaddr,
			  cdd->mem->region.desc_bus_address[0]);
	kfree(cmd);
}

static struct class csx_mem_class = {
	.name = "csxmem",
	.release = csx_mem_class_release,
};

/*********************
 * Interrupt handler *
 *********************/

/* Determines which type(s) of interrupt were triggered, and wakes
 * up the appropriate waiters.
 */
static irqreturn_t
csx_intr_handler(int irq, void *arg, struct pt_regs *regs)
{
	struct csx_device_data *cdd;
	uint32_t pcistatus, dmastatus;
	irqreturn_t ret = IRQ_NONE;

	cdd = (struct csx_device_data *) arg;

	pcistatus = readl(&cdd->ctl->regs[CSX_CTL_PCIINTR_INDX]);
	dmastatus = readl(&cdd->ctl->regs[CSX_CTL_DMAINTR_INDX]);

	if (pcistatus & CSX_CTL_PCIINTR_MASK) {
		uint32_t pcimask;
		atomic_inc(&cdd->ctl->intr_count);
		// unmask the interrupt - user code will sort it later
		writel(0, &cdd->ctl->regs[CSX_CTL_PCIINTRMASK_INDX]);
		MMIOWB();
		// need some sort of memory barrier on x86 to avoid spurious interrupts
		pcimask = readl(&cdd->ctl->regs[CSX_CTL_PCIINTRMASK_INDX]);
		wake_up_all(&cdd->ctl->intr_queue);
		ret = IRQ_HANDLED;
	}
	if (dmastatus & CSX_CTL_DMAINTR_MASK) {
		atomic_inc(&cdd->mem->intr_count);
		/* Ack the interrupt, leave the mask alone */
		// dma is done by kernel so this is different from pci interrupt
		writel(dmastatus, &cdd->ctl->regs[CSX_CTL_DMAINTR_INDX]);
		MMIOWB();
		// need some sort of memory barrier on x86 to avoid spurious interrupts
		dmastatus = readl(&cdd->ctl->regs[CSX_CTL_DMAINTR_INDX]);
		wake_up_all(&cdd->mem->intr_queue);
		ret = IRQ_HANDLED;
	}

	return ret;
}

/****************
 * Entry points *
 ****************/

/* Control register operations */

/* Seek to a particular position in the control register aperture */
static loff_t
csx_ctl_llseek(struct file *filp, loff_t off, int whence)
{
	struct csx_ctl_data *ccd;
	loff_t pos;

	ccd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_ctl_data, cdev);

	switch (whence) {
		case 0:	/* SEEK_SET */
			pos = off;
			break;
		case 1:	/* SEEK_CUR */
			pos = filp->f_pos + off;
			break;
		case 2:	/* SEEK_END */
			pos = ccd->len + off;
			break;
		default:
			return -EINVAL;
	}

	/* Verify seek does not land outside control register area */
	if (pos < 0 || pos > ccd->len)
		return -EINVAL;
	filp->f_pos = pos;
	return pos;
}

/* Memory-map the control register aperture
 * XXX: On SGI Altix, the user mapping is unreliable before Linux 2.6.17.
 */
static int
csx_ctl_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct csx_ctl_data *ccd;
	unsigned long off;
	unsigned long vsize;
	int ret;

	ccd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_ctl_data, cdev);

	/* Verify request does not extend past end of control register
	 * aperture.
	 */
	off = vma->vm_pgoff << PAGE_SHIFT;
	vsize = vma->vm_end - vma->vm_start;
	if (off + vsize > ccd->len)
		return -EINVAL;

	/* Ensure mapping is suitable for IO */
	vma->vm_flags |= VM_IO | VM_RESERVED;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

#if LINUX_VERSION_CODE <=  KERNEL_VERSION(2,6,12) 	 
    ret = io_remap_page_range(vma, vma->vm_start,
        ccd->bar + off,
        vsize, vma->vm_page_prot);
#else 	 
     ret = remap_pfn_range(vma, vma->vm_start, (unsigned long)(ccd->bar+off)>>PAGE_SHIFT, 	 
        vsize, vma->vm_page_prot); 	 
#endif 	 
 
	if (ret)
		return -EAGAIN;

	return 0;
}

/* Open the control register aperture */
static int
csx_ctl_open(struct inode *inode, struct file *filp)
{
	struct csx_ctl_data *ccd;
	struct csx_ctl_file *ccf;

	ccd = container_of(inode->i_cdev, struct csx_ctl_data, cdev);
#if CSX_DEBUG
	printk(KERN_DEBUG "%s: ccd %p\n", __FUNCTION__, ccd);
#endif

	ccf = kmalloc(sizeof(struct csx_ctl_file), GFP_KERNEL);
	if (NULL == ccf)
		return -ENOMEM;

	filp->private_data = ccf;
	atomic_set(&ccd->intr_count, 0);
	ccf->intr_count = atomic_read(&ccd->intr_count);
	/* Mask MTAP interrupts */
	writel(0, &ccd->regs[CSX_CTL_PCIINTRMASK_INDX]);
	MMIOWB();

	return 0;
}

/* Wait for a PCI interrupt (i.e. FPGA program or status/fault interrupt) */
/*
	the pci interrupt mask must be opened in user code before the poll
	- this is a different strategy to the dma interrupt
 */
static unsigned int
csx_ctl_poll(struct file *filp, struct poll_table_struct *polltab)
{
	struct csx_ctl_data *ccd;
	struct csx_ctl_file *ccf;
	__s32 sample;

	ccd = container_of(filp->f_dentry->d_inode->i_cdev,
		struct csx_ctl_data, cdev);
	ccf = filp->private_data;

	poll_wait(filp, &ccd->intr_queue, polltab);

	/* CSX poll implements a wait for interrupt */
	sample = atomic_read(&ccd->intr_count);
	if (ccf->intr_count != sample)
	{
		ccf->intr_count = sample;
		return (POLLIN | POLLRDNORM);
	}
	return 0;
}

/* Read control register */
static ssize_t
csx_ctl_read(struct file *filp, char __user * data, size_t size, loff_t * off)
{
	uint32_t val;
	struct csx_ctl_data *ccd;
	size_t index;

	ccd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_ctl_data, cdev);

	/* Verify size and offset are correct multiples of register size */
	if ((size & (CSX_CTL_REGSIZE - 1)) || (*off & (CSX_CTL_REGSIZE - 1)))
		return -EINVAL;

	/* Verify access doesn't fall outside control register area */
	if ((*off < 0) || (*off + size > ccd->len))
		return -EINVAL;

	for (index = 0; index < size / CSX_CTL_REGSIZE;
	     index += CSX_CTL_REGSIZE) {
		/* Using intermediate value allows unaligned user data, and
		 * checks against user space pointer
		 */
		val = readl((void *)((size_t)(ccd->regs) + (size_t)*off + index));
             
		if (unlikely
		    (copy_to_user(&data[index], &val, CSX_CTL_REGSIZE)))
			return -EFAULT;
	}

	*off += size;
	return size;
}

/* Release per-file resources for control register aperture */
static int
csx_ctl_release(struct inode *inode, struct file *filp)
{
	struct csx_ctl_file *ccf;
	struct csx_ctl_data *ccd;

	ccd = container_of(inode->i_cdev, struct csx_ctl_data, cdev);
	// often closes with mask open
	/* close pci interrupts*/
	writel(0, &ccd->regs[CSX_CTL_PCIINTRMASK_INDX]);
	MMIOWB();

	ccf = filp->private_data;
	kfree(ccf);

	return 0;
}

/* Write control register */
static ssize_t
csx_ctl_write(struct file *filp, const char __user * data, size_t size,
	      loff_t * off)
{
	uint32_t val;
	struct csx_ctl_data *ccd;
	size_t index;

	ccd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_ctl_data, cdev);

	/* Verify size and offset are correct multiples of register size */
	if ((size & (CSX_CTL_REGSIZE - 1)) || (*off & (CSX_CTL_REGSIZE - 1)))
		return -EINVAL;

	/* Verify access doesn't fall outside control register area */
	if ((*off < 0) || (*off + size > ccd->len))
		return -EINVAL;

	for (index = 0; index < size / CSX_CTL_REGSIZE;
	     index += CSX_CTL_REGSIZE) {
		/* Using intermediate value allows unaligned user data,
		 * and checks against user space pointer
		 */
		if (unlikely(copy_from_user(&val, &data[index],
					    CSX_CTL_REGSIZE))) {
			MMIOWB();
			return -EFAULT;
		}
		writel(val, (void *)((size_t)(ccd->regs) + (size_t)*off + index));
	}

	/* Ensure writes arrive at hardware before proceeding */
	MMIOWB();

	*off += size;
	return size;
}

struct file_operations csx_ctl_fops = {
	.owner = THIS_MODULE,
	.llseek = csx_ctl_llseek,
	.mmap = csx_ctl_mmap,
	.open = csx_ctl_open,
	.poll = csx_ctl_poll,
	.read = csx_ctl_read,
	.release = csx_ctl_release,
	.write = csx_ctl_write,
};


/* Memory operations */

/* DMA routines */

/* On buffer locking, we need to create a scatter/gather list and other
 * information.  We cannot rely on the user giving us back the required
 * information so we record this in a structure in the driver.
 */

static int
sgl_map_user_pages(struct scatterlist *sgl, const unsigned int max_pages,
		   unsigned long user_addr, size_t count,
		   enum dma_data_direction direction)
{
	int res, i, j;
	struct page **pages;
	const unsigned long end =
		(user_addr + count + PAGE_SIZE - 1) >> PAGE_SHIFT;
	const unsigned long start = user_addr >> PAGE_SHIFT;
	const int nr_pages = end - start;
	const int read_from_board = (direction == DMA_FROM_DEVICE ? 1 : 0);

	if( sgl == NULL )
	{
		return -EINVAL;
	}

	/* User attempted Overflow! */
	if ((user_addr + count) < user_addr)
	{
		return -EINVAL;
	}

	/* Too big */
	if (nr_pages > max_pages)
	{
		return -ENOMEM;
	}

	/* Hmm? */
	if (count == 0)
	{
		return 0;
	}

	// This memory allocation doesn't live beyond the end of this function no specific cleanup needed.
	if ((pages = kmalloc(max_pages * sizeof(*pages), GFP_KERNEL)) == NULL)
	{
		return -ENOMEM;
	}

	memset(pages, 0, max_pages * sizeof(*pages));

#if CSX_DEBUG
	printk(KERN_DEBUG "%s: calling get_user_pages ua:%p, nr_pages %d, %s max_pages=%d\n",
		__FUNCTION__, (void *) user_addr, nr_pages, (read_from_board ? "read" : "write"), max_pages);
#endif

	/* Try to fault in all of the necessary pages */
	down_read(&current->mm->mmap_sem);

	/* read_from_board means reading from board */
	// Which means write to memory = 1
	res = get_user_pages(current, current->mm, user_addr, nr_pages, read_from_board, 0, pages, NULL);

	up_read(&current->mm->mmap_sem);

#if CSX_DEBUG
	printk(KERN_DEBUG "%s: get_user_pages return %d\n", __FUNCTION__, res);
#endif

	/* Errors and no page mapped should return here */
	if (res < nr_pages)
	{
        goto out_unmap; /* Consider this an error */
	}

//	for (i = 0; i < nr_pages; i++)
//	{
		/* FIXME: flush superflous for rw==READ,
		 * probably wrong function for rw==WRITE
		 */
		/*     flush_dcache_page(pages[i]); QQQ */
//	}

	/* Populate the scatter/gather list */
	sgl[0].page = pages[0];
	sgl[0].offset = user_addr & ~PAGE_MASK;
	if (nr_pages > 1)
	{
		sgl[0].length = PAGE_SIZE - sgl[0].offset;
		count -= sgl[0].length;

		for (i = 1; i < nr_pages; i++)
		{
			sgl[i].page = pages[i];
			sgl[i].offset = 0;
			sgl[i].length = count < PAGE_SIZE ? count : PAGE_SIZE;
			count -= PAGE_SIZE;

#if CSX_DEBUG
			if( (count < 0) && (i != (nr_pages-1)) )
			{
				printk(KERN_DEBUG "%s: ASSERT FAILURE: count going negaticve with more pages to go ?!\n", __FUNCTION__);
			}
#endif
		}
	}
	else
	{
		sgl[0].length = count;

#if CSX_DEBUG
		if( (sgl[0].offset + sgl[0].length) > PAGE_SIZE )
		{
			printk(KERN_DEBUG "%s: ASSERT FAILURE: sg offset + size > page_size ?!\n", __FUNCTION__);
		}
#endif
	}

	kfree(pages);
	return nr_pages;

out_unmap:
	if (res > 0)
	{
		for (j = 0; j < res; j++)
		{
			page_cache_release(pages[j]);
		}
		res = 0;
	}
	kfree(pages);
	return -ENOMEM; /* We consider this a fail */
}

static int
sgl_unmap_user_pages(struct scatterlist *sgl, const unsigned int nr_pages,
		     int dirtied)
{
	int i;

	for (i = 0; i < nr_pages; i++)
	{
		struct page *page = sgl[i].page;

#if CSX_DEBUG
		if( page == NULL )
		{
			printk(KERN_DEBUG "%s: ASSERT FAILURE: page is NULL for sg[%d] ?!\n", __FUNCTION__, i);
		}
#endif

		if (dirtied)
		{
			SetPageDirty(page);
		}
		/* FIXME: cache flush missing for rw==READ
		 * FIXME: call the correct reference counting function
		 */
		page_cache_release(page);
	}

	return 0;
}

/* Pin down user pages and put them into a scatter gather list.
 * Returns <= 0 if:
 * - mapping of all pages not successful
 * - any page is above max_pfn
 * (i.e., either completely successful or fails)
 */
static int
csx_map_user_pages(struct scatterlist *sgl,
		   const unsigned int max_pages,
		   unsigned long user_addr, size_t count,
		   enum dma_data_direction dma_direction,
		   unsigned long max_pfn)
{
	int nr_pages;

#if CSX_DEBUG
	printk(KERN_DEBUG "%s called\n", __FUNCTION__);
#endif

	nr_pages = sgl_map_user_pages(sgl, max_pages, user_addr, count, dma_direction);

#if CSX_DEBUG
	printk(KERN_DEBUG "%s sgl_map_user_pages returned %d for %x\n", __FUNCTION__, nr_pages, (unsigned) count);
#endif
	if (nr_pages <= 0)
	{
		return nr_pages;	/* no cleanup required. */
	}

	/* QQQ Not sure about the max_pfn piece. */
#ifdef MAX_PFN
	for (i = 0; i < nr_pages; i++)
	{
		if (page_to_pfn(sgl[i].page) > max_pfn)
		{
			goto out_unmap;
		}
	}
#endif
	return nr_pages;

#ifdef MAX_PFN
out_unmap:
#endif

	sgl_unmap_user_pages(sgl, nr_pages, 0);
	return -EINVAL;
}

/* Take the scatterlist and map to the IOMMU */
static int
csx_map_to_iommu(struct pci_dev *pdev, struct scatterlist *sgl,
		 unsigned pages_pinned, struct dma_desc *desc_array,
		 enum dma_data_direction direction)
{
	int pages_mapped;
	int i;

	/* map to bus */
#if CSX_DEBUG
//	printk(KERN_DEBUG "%s: requesting the following mappings from dma_map_sg:\n", __FUNCTION__);
//	for (i = 0; i < pages_pinned; i++)
//	{
//		printk(KERN_DEBUG "\tsgl[%d] page=%p offset=%u length=%u dma_address=%llx dma_length=%u\n",
//		       i, sgl[i].page, sgl[i].offset, sgl[i].length,  sg_dma_address(&sgl[i]), sg_dma_len(&sgl[i]));
//	}
#endif
	pages_mapped = dma_map_sg(&pdev->dev, sgl, pages_pinned, direction);

	if (pages_mapped == 0) {
#if CSX_DEBUG
		printk(KERN_DEBUG "%s: dma_map_sg returned %d\n",
		       __FUNCTION__, pages_mapped);
#endif
		return -ENOSPC;
	}

	/* Now copy out to user's desc_array buffer. */
	/* This check should be unecessary but this is the point at which we write to the
	   desc_arrary so we should just check that it's big enough */
	if (pages_mapped > MAX_SG_SIZE) {
#if CSX_DEBUG
	  /* This is really an internal error. */
	  printk(KERN_DEBUG "%s: pages_mapped (%d) larger than max (%d)\n",
                 __FUNCTION__, pages_mapped, (int)MAX_SG_SIZE);
#endif
	  return -ENOSPC;
	}

	for (i = 0; i < pages_mapped; i++)
	{
		/* QQQ what about the offset field.  */
	    desc_array[i].bus_address = (bus_address_t)sg_dma_address(&sgl[i]);
		desc_array[i].length = sg_dma_len(&sgl[i]);

        if( desc_array[i].bus_address == 0 )
        {
            printk(KERN_DEBUG "WARNING: page descriptor %d of %d has zero bus address.\n",
                i, pages_mapped );
        }

#if CSX_DEBUG
//		printk(KERN_DEBUG "%s:%d: %llx %x\n", __FUNCTION__, i,
//		       desc_array[i].bus_address,
//		       desc_array[i].length);
#endif
	}

	return pages_mapped;	/* which may be different from pages_pinned */
}

/* Data buffer lock function, specification as user call */
static status
csx_lock_buffer_for_dma(struct csx_device_data *cdd,
            int buffer,
            void *user_address,
			unsigned user_size, int *desc_count,
			struct dma_desc *desc_array, unsigned flags)
{
	int pages_pinned = 0;
	int pages_mapped = 0;
	const unsigned long max_pfn = 4000;	/*QQQ */
	status ret;
	char *errstring = "";
	enum dma_data_direction dma_direction;

#if CSX_DEBUG
	printk(KERN_DEBUG "%s called\n", __FUNCTION__);
#endif

    if( buffer < 0 || buffer > 1 )
    {
  		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, buffer );
		return -EINVAL;
    }

	/* Allocate an initialise the dma_region
	 * Freed in csx_unlock_buffer_for_dma
	 */
	cdd->mem->region.sglist[buffer] =
		kmalloc(MAX_SG_SIZE * sizeof(struct scatterlist), GFP_KERNEL);

	if (!cdd->mem->region.sglist[buffer]) {
		errstring = "Allocation of scatter/gather failed";
		ret = -ENOMEM;
		goto alloc_sglist_failed;
	}

	/* Clear the scatterlist */
	memset(cdd->mem->region.sglist[buffer], 0,
	       MAX_SG_SIZE * sizeof(struct scatterlist));

	/* Translate from private flags to DMA enum types.  */
	if (flags & DMA_FLAG_FROM_DEVICE)
	{
		dma_direction = DMA_FROM_DEVICE;
	}
	else
	{
		dma_direction = DMA_TO_DEVICE;
	}

	/* Pin the pages */
	pages_pinned =
		csx_map_user_pages(cdd->mem->region.sglist[buffer], MAX_SG_SIZE,
				   (unsigned long) user_address, user_size,
				   dma_direction, max_pfn);

    if (pages_pinned <= 0) {
		errstring = "csx_map_user_pages failed";
		ret = (status) pages_pinned;
		goto pin_failed;
	}

    /* Pages pinned, now map to bus addresses */
	pages_mapped = csx_map_to_iommu(cdd->pcidev, cdd->mem->region.sglist[buffer],
					pages_pinned, desc_array,
					dma_direction);
	if (pages_mapped <= 0) {
		ret = (status) pages_mapped;
		goto iomapping_failed;
	}

	/* We assume that this buffer will only be used by the device from now on */
	// ceb. 6.3.06 Perhaps only if dma_direction == Write. 
	dma_sync_sg_for_device(&cdd->pcidev->dev, cdd->mem->region.sglist[buffer],
			       pages_pinned, dma_direction);

	/* Successfully programmed the IOMMU
	 * Gather up results and pass back to user side caller.
	 */
	*desc_count = pages_mapped;

	/* Record actual pages_pinned. We need this rather than pages_mapped which
	 * is what pci_map_sg returns.
	 */
	cdd->mem->region.pages_pinned[buffer] = pages_pinned;

	return 0;		/* success */

iomapping_failed:
pin_failed:
alloc_sglist_failed:
	printk(KERN_WARNING "%s: %s (%d)\n", __FUNCTION__, errstring,
	       pages_mapped);
	return ret;
}

static int
csx_unmap_to_iommu(struct pci_dev *pdev, struct scatterlist *sgl,
		   unsigned pages_mapped, enum dma_data_direction direction)
{
	int i;

	dma_unmap_sg(&pdev->dev, sgl, pages_mapped, direction);

	/* This is essentially put_user_pages */
	for (i = 0; i < pages_mapped; i++)
	{
		if (direction == DMA_FROM_DEVICE)
		{
			if (!PageReserved(sgl[i].page))
			{
				SetPageDirty(sgl[i].page);
			}
		}
		page_cache_release(sgl[i].page);
	}
	return 0;		/* can't fail I guess. */
}

static status
csx_unlock_buffer_for_dma(struct csx_device_data *cdd,
    int buffer,
    int pages_mapped,
    struct dma_desc *desc_array,
    unsigned flags)
{
	int reso;
	enum dma_data_direction direction;

    if( buffer < 0 || buffer > 1 )
    {
  		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, buffer );
		return -EINVAL;
    }

	if (flags & DMA_FLAG_FROM_DEVICE)
	{
		direction = DMA_FROM_DEVICE;
	}
	else
	{
		direction = DMA_TO_DEVICE;
	}

	reso = csx_unmap_to_iommu(cdd->pcidev, cdd->mem->region.sglist[buffer],
				  pages_mapped, direction);

	/* If we have read something from the device we need to
	   synch to allow the cpu to read it. */
	if (direction == DMA_FROM_DEVICE)
	  dma_sync_sg_for_cpu(&cdd->pcidev->dev, cdd->mem->region.sglist[buffer],
			       pages_mapped, direction);

	/* Now release the scatter/gather list */
	kfree(cdd->mem->region.sglist[buffer]);
	cdd->mem->region.sglist[buffer] = NULL;

	return reso;
}

#define DMA_PCI_ADDRESS_SPACE 0x2

/* Set this for pur chain mode - else it will copy first descriptor in */
#define FIRE_DMA_PURE_CHAIN   0

static int
csx_fire_dma(struct csx_device_data *cdd, int buffer, bus_address_t bus_addr,
	     unsigned flags)
{
	int reso;

#if !FIRE_DMA_PURE_CHAIN
    unsigned int n;
    unsigned int byte_count = 0;
    unsigned int *p_desc;
#endif

    if( buffer < 0 || buffer > 1 )
    {
  		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, buffer );
		return -EINVAL;
    }

    if (bus_addr == cdd->mem->region.desc_bus_address[buffer])
	{
        /* call the sync function to ensure that the data buffer is ready for
		 * the device
		 */
		dma_sync_sg_for_device(&cdd->pcidev->dev,
				       cdd->mem->region.sglist[buffer],
				       cdd->mem->region.n_pages[buffer],
				       cdd->mem->region.direction);

#if FIRE_DMA_PURE_CHAIN
		/* Say the address is in PCI space.  */
		bus_addr |= DMA_PCI_ADDRESS_SPACE;

		/* Set up for Pure chain & zero byte count */
		writel( 0, &cdd->ctl->regs[CSX_DMA_BYTECOUNT_INDX]);
		writel( 0x00c00000, &cdd->ctl->regs[CSX_DMA_CONTROL_INDX]);

#if __LP64__
		writeq( 0, &cdd->ctl->regs[CSX_DMA_FPGA_ADDR]);
		writeq( 0, &cdd->ctl->regs[CSX_DMA_PCI_ADDR]);
		writeq( bus_addr, &cdd->ctl->regs[CSX_DMA_DESC_INDX]);
#else
		writel(0, &cdd->ctl->regs[CSX_DMA_FPGA_ADDR]);
		writel(0, &cdd->ctl->regs[CSX_DMA_FPGA_ADDR + 1]);
		writel(0, &cdd->ctl->regs[CSX_DMA_PCI_ADDR]);
		writel(0, &cdd->ctl->regs[CSX_DMA_PCI_ADDR + 1]);
		writel(bus_addr & 0xffffffff, &cdd->ctl->regs[CSX_DMA_DESC_INDX]);
		writel(bus_addr >> 32, &cdd->ctl->regs[CSX_DMA_DESC_INDX + 1]);
#endif

		/* Finally, Fire DMA by setting the valid bit */
		/* Probably a good ideal to request a memory barrier here. */
		wmb();
		writel( 0x01000000, &cdd->ctl->regs[CSX_DMA_BYTECOUNT_INDX]);
		MMIOWB();
#else
        if (buffer == 0)
        {
            p_desc = (unsigned int *)cdd->mem->region.desc_kaddr;
        }
        else
        {
            p_desc = (unsigned int *)( (char *)cdd->mem->region.desc_kaddr + MAX_DESCRIPTOR_BUFFER_SIZE );
        }
       
        for( n=0; n<(32/4); n++ )
        {
            if( n == 0 )
            {
                /* Strip off valid bit for now */
                byte_count = *p_desc & 0xFFFFFF;
            }
            else
            {
        		writel( *p_desc, &cdd->ctl->regs[CSX_DMA_BYTECOUNT_INDX + n]);
            }

            p_desc++;
        }

		/* Finally, Fire DMA by or-ing in the valid bit */
		wmb();
		writel( byte_count | 0x01000000, &cdd->ctl->regs[CSX_DMA_BYTECOUNT_INDX]);
		MMIOWB();
#endif
        reso = 0;
	}
	else
	{
		reso = -EINVAL;
	}

	return reso;
}


/* End of DMA routines */


static int
csx_mem_ioctl(struct inode *inode, struct file *file, unsigned int cmd_in,
	      unsigned long arg)
{
	struct csx_mem_data *cmd;
	struct csx_device_data *cdd;
	void __user *p = (void __user *) arg;

	cmd = container_of(file->f_dentry->d_inode->i_cdev, struct csx_mem_data, cdev);
	cdd = cmd->maindata;

	switch (cmd_in)
    {
	case CSX_GET_COHERENT_BUFFER_INFO:
	{
		struct ioctl_param_get_coherent_map_info this_call;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_GET_COHERENT_BUFFER_INFO:\n" );
#endif
		memset(&this_call, 0, sizeof this_call);

		/* No input parameters.  */
		this_call.descriptor_chain1 = cdd->mem->region.desc_bus_address[0];
		this_call.descriptor_chain2 = cdd->mem->region.desc_bus_address[1];
		this_call.bounce_buffer = cdd->mem->region.data_bus_address;
		this_call.bounce_buffer_size = cdd->mem->region.data_buf_size;

		//printk(KERN_DEBUG "%s: about to copy to user p = %p\n", __FUNCTION__, p);
		if (copy_to_user(p, &this_call, sizeof this_call) != 0)
			return -EFAULT;
		return 0;
	}

	case CSX_COPY_TO_DESC_BUFFER:
	{
		struct ioctl_param_copy_to_coherent_descriptor_buffer this_call;
		int reso;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_COPY_TO_DESC_BUFFER:\n" );
#endif

		if( copy_from_user(&this_call, p, sizeof this_call) != 0 )
        {
			return -EFAULT;
        }

        if( this_call.buffer < 0 || this_call.buffer > 1 )
        {
    		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, this_call.buffer );
			return -EINVAL;
        }

		/* check size */
		if( (this_call.size == 0) ||
			(this_call.size > cdd->mem->region.desc_buf_size) )
        {
			reso = -EINVAL;
        }
		else
        {
            unsigned long copy_ret;

			/* safe to do the copy */
            if( this_call.buffer == 0 && this_call.desc_address == cdd->mem->region.desc_bus_address[0] )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_TO_DESC_BUFFER: copy to buffer A.\n", __FUNCTION__ );
#endif
                copy_ret = copy_from_user( cdd->mem->region.desc_kaddr,
					this_call.user_descriptor_buffer, this_call.size );
            }
            else if( this_call.buffer == 1 && this_call.desc_address == cdd->mem->region.desc_bus_address[1] )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_TO_DESC_BUFFER: copy to buffer B.\n", __FUNCTION__ );
#endif
                copy_ret = copy_from_user( (char *)cdd->mem->region.desc_kaddr +
                    MAX_DESCRIPTOR_BUFFER_SIZE,
					this_call.user_descriptor_buffer, this_call.size );
            }
            else
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_TO_DESC_BUFFER: bad copy location.\n", __FUNCTION__ );
#endif
				return -EINVAL;
            }

			if( copy_ret != 0 )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_TO_DESC_BUFFER: copy_from_user failed.\n", __FUNCTION__ );
#endif
				return -EINVAL;
            }

#if CSX_DEBUG_EXTREME
			{
				unsigned int *ptr =
					(unsigned int *) cdd->mem->
					region.desc_kaddr;
				int i;

				for (i = 0;
					i <
					this_call.size /
					(sizeof(unsigned int)); i++) {
					printk(KERN_DEBUG "%x: %x\n",
						i, *ptr);
					ptr++;
				}
			}
#endif
			reso = 0;
		}
		return reso;
	}

	case CSX_COPY_DATA_BUFFER:
	{
		struct ioctl_param_copy_data_buffer this_call;
		int reso = 0;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_COPY_DATA_BUFFER:\n" );
#endif

		if (copy_from_user(&this_call, p, sizeof this_call) != 0)
			return -EFAULT;

		/* some checks. */
		if (!cdd->mem->region.data_kaddr)
			return -EINVAL;
		if ((this_call.size == 0) ||
			(this_call.size > cdd->mem->region.data_buf_size))
			return -EINVAL;

		/* Copy to or from */
		switch (this_call.flags) {
		case DMA_FLAG_TO_DEVICE:
			/* copy from user to kernel */
			if (copy_from_user(cdd->mem->region.data_kaddr,
					this_call.user_data_buffer,
					this_call.size) != 0)
				return -EFAULT;
			break;
		case DMA_FLAG_FROM_DEVICE:
			/* copy kernel to user */
			if (copy_to_user(this_call.user_data_buffer,
					cdd->mem->region.data_kaddr,
					this_call.size) != 0)
				return -EFAULT;

			break;
		default:
			reso = -EINVAL;
		}
		return reso;
	}

	case CSX_LOCK_DATA_BUFFER:
	{
		struct ioctl_param_buffer_lock this_call;
		struct dma_desc *desc_array;
		int reso;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_LOCK_DATA_BUFFER:\n" );
#endif

		desc_array = (struct dma_desc*)kmalloc(
			MAX_SG_SIZE * sizeof(struct dma_desc),
			GFP_KERNEL);

		if (NULL == desc_array)
        {
            return -ENOMEM;
        }

		if (copy_from_user(&this_call, p, sizeof this_call) != 0)
        {
            kfree( desc_array );
            return -EFAULT;
        }

        if( this_call.buffer < 0 || this_call.buffer > 1 )
        {
            printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, this_call.buffer );
            kfree( desc_array );
            return -EINVAL;
        }

#if CSX_DEBUG
		printk(KERN_DEBUG
			"%s: csx_lock_data_buffer for %x bytes\n",
			__FUNCTION__, this_call.user_size);
#endif

		/* a few checks */
		if (this_call.user_size > MAX_SINGLE_DMA_TRANSFER)
        {
            kfree( desc_array );
            return -EINVAL;
		}

		/* clear the desc_array */
		memset(desc_array, 0, MAX_SG_SIZE * sizeof(struct dma_desc));

		reso = csx_lock_buffer_for_dma(cdd,
			this_call.buffer,
			this_call.user_address,
			this_call.user_size,
			&this_call.desc_count,
			desc_array,
			this_call.flags);

		if (!reso)
		{
		    /* Copy the data buffer back. */
		    if (copy_to_user(this_call.desc_array,
                desc_array,
                this_call.desc_count *
                sizeof(struct dma_desc)) != 0)
            {
                kfree( desc_array );
                return -EFAULT;
            }

            if (copy_to_user(p, &this_call,
                sizeof(struct
                ioctl_param_buffer_lock)) != 0)
            {
                kfree( desc_array );
                return -EFAULT;
            }
		}

		kfree( desc_array );
		return reso;
	}

	case CSX_UNLOCK_DATA_BUFFER:
	{
		struct ioctl_param_buffer_unlock this_call;
		struct dma_desc *desc_array;
		int reso;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_UNLOCK_DATA_BUFFER:\n" );
#endif

		if (copy_from_user(&this_call, p, sizeof this_call) != 0)
			return -EFAULT;

        if( this_call.buffer < 0 || this_call.buffer > 1 )
        {
    		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, this_call.buffer );
			return -EINVAL;
        }

#if CSX_DEBUG
		printk(KERN_DEBUG
			"%s: csx_unlock_data_buffer for %d descriptors\n",
			__FUNCTION__, this_call.desc_count);
#endif

		/* How do we check that this unlock corresponds to an outstanding lock??
		 * check if pages_pinned is sensible compared to what the user has given us
		 * pages_pinned can be the same as or less than what we're given back as the 
		 * putative size. 
		 */
		if (cdd->mem->region.pages_pinned[this_call.buffer] < this_call.desc_count)
			return -EINVAL;

		/* We have to use the pages pinned number rather then the desc_count that we delivered
		 * to the user
		 */
		if (!cdd->mem->region.pages_pinned[this_call.buffer])
			return -EINVAL;

		desc_array = (struct dma_desc*)kmalloc(
			MAX_SG_SIZE * sizeof(struct dma_desc),
			GFP_KERNEL);
		if (NULL == desc_array)
			return -ENOMEM;

		/* clear the desc_array */
		memset(desc_array, 0, MAX_SG_SIZE * sizeof(struct dma_desc));

		reso = csx_unlock_buffer_for_dma(cdd,
            this_call.buffer,
			cdd->mem->region.pages_pinned[this_call.buffer],
			desc_array,
			this_call.flags);

		/* Use this to prevent dangling releases.  */
		cdd->mem->region.pages_pinned[this_call.buffer] = 0;

		kfree(desc_array);

		return reso;
	}

	case CSX_FIRE_DMA:
	{
		struct ioctl_param_buffer_fire this_call;
		int reso;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_FIRE_DMA:\n" );
#endif

		if (copy_from_user(&this_call, p, sizeof this_call) != 0)
			return -EFAULT;

        if( this_call.buffer < 0 || this_call.buffer > 1 )
        {
    		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, this_call.buffer );
			return -EINVAL;
        }

#if CSX_DEBUG
		printk(KERN_DEBUG
			"%s: csx_fire_dma for desc address %llx\n",
			__FUNCTION__, this_call.desc_address);
#endif
		reso = csx_fire_dma(cdd, this_call.buffer, this_call.desc_address,
			this_call.flags);

		return reso;
	}

    case CSX_COPY_AND_FIRE:
    {
		struct ioctl_param_copy_and_fire this_call;
		int reso;

#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "CSX_COPY_AND_FIRE:\n" );
#endif

		if( copy_from_user(&this_call, p, sizeof this_call ) != 0 )
        {
#if CSX_DEBUG
    		printk(KERN_DEBUG "%s: copy_from_user failed\n", __FUNCTION__ );
#endif
			return -EFAULT;
        }

        if( this_call.buffer < 0 || this_call.buffer > 1 )
        {
    		printk(KERN_WARNING "%s: buffer is out of range (%d)\n", __FUNCTION__, this_call.buffer );
			return -EINVAL;
        }

		/* check size */
		if ((this_call.size == 0) ||
			(this_call.size > cdd->mem->region.desc_buf_size))
        {
#if CSX_DEBUG
    		printk(KERN_DEBUG "%s: size error\n", __FUNCTION__ );
#endif
			reso = -EINVAL;
        }
		else
        {
            unsigned long copy_ret;

            /* safe to do the copy */
            if( this_call.buffer == 0 && this_call.desc_address == cdd->mem->region.desc_bus_address[0] )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_AND_FIRE: copy to buffer A.\n", __FUNCTION__ );
#endif
                
                copy_ret = copy_from_user( cdd->mem->region.desc_kaddr,
					this_call.user_descriptor_buffer, this_call.size );
            }
            else if( this_call.buffer == 1 && this_call.desc_address == cdd->mem->region.desc_bus_address[1] )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_AND_FIRE: copy to buffer B.\n", __FUNCTION__ );
#endif
                
                copy_ret = copy_from_user( (char *)cdd->mem->region.desc_kaddr +
                    MAX_DESCRIPTOR_BUFFER_SIZE,
					this_call.user_descriptor_buffer, this_call.size );
            }
            else
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_AND_FIRE: bad copy location.\n", __FUNCTION__ );
#endif
				return -EINVAL;
            }

			if( copy_ret != 0 )
            {
#if CSX_DEBUG
        		printk(KERN_DEBUG "%s: CSX_COPY_AND_FIRE: copy_from_user failed.\n", __FUNCTION__ );
#endif
				return -EINVAL;
            }
            else
            {
  		        reso = csx_fire_dma( cdd, this_call.buffer, this_call.desc_address, this_call.flags );
                if( reso < 0 )
                {
            		printk(KERN_DEBUG "%s: csx_fire_dma failed.\n", __FUNCTION__ );
                }
            }
		}

		return reso;
    }

    case CSX_GET_PAGE_SIZE:
    {
		struct ioctl_param_get_page_size this_call;

        this_call.page_size_in_bytes = PAGE_SIZE;

        if (copy_to_user(p, &this_call, sizeof this_call) != 0)
        {
			return -EFAULT;
        }
        else
        {
            return 0;
        }
    }

	default:
#if CSX_DEBUG_IOCTLS
        printk(KERN_DEBUG "Bad ioctl command !\n" );
#endif
		return -ENOTTY;
	}
}

/* Seek to a particular position in the memory aperture */
static loff_t
csx_mem_llseek(struct file *filp, loff_t off, int whence)
{
	struct csx_mem_data *cmd;
	loff_t pos;

	cmd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_mem_data, cdev);

	switch (whence) {
		case 0:	/* SEEK_SET */
			pos = off;
			break;
		case 1:	/* SEEK_CUR */
			pos = filp->f_pos + off;
			break;
		case 2:	/* SEEK_END */
			pos = cmd->len + off;
			break;
		default:
			return -EINVAL;
	}

	/* Verify seek does not land outside memory area */
	if (pos < 0 || pos > cmd->len)
		return -EINVAL;
	filp->f_pos = pos;
	return pos;
}

/* Memory-map the memory aperture.
 * XXX: On SGI Altix, the user mapping is unreliable before Linux 2.6.17.
 */
static int
csx_mem_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct csx_mem_data *cmd;
	unsigned long off;
	unsigned long vsize;
	int ret;

	cmd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_mem_data, cdev);

	/* Verify request does not extend past end of memory aperture. */
	off = vma->vm_pgoff << PAGE_SHIFT;
	vsize = vma->vm_end - vma->vm_start;
	if (off + vsize > cmd->len)
		return -EINVAL;

	/* Ensure mapping is suitable for IO */
	vma->vm_flags |= VM_IO | VM_RESERVED;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

#if LINUX_VERSION_CODE <=  KERNEL_VERSION(2,6,12) 	 
    ret = io_remap_page_range(vma, vma->vm_start,
        cmd->bar + off,
        vsize, vma->vm_page_prot);
#else 	 
    ret = remap_pfn_range(vma, vma->vm_start, ((unsigned long) cmd->bar+off)>>PAGE_SHIFT, 	 
        vsize, vma->vm_page_prot); 	 
#endif 	 

	return ret;
}

/* Open the memory aperture */
static int
csx_mem_open(struct inode *inode, struct file *filp)
{
	struct csx_mem_data *cmd;
	struct csx_mem_file *cmf;

	cmd = container_of(inode->i_cdev, struct csx_mem_data, cdev);
#if CSX_DEBUG
	printk(KERN_DEBUG "%s: cmd %p\n", __FUNCTION__, cmd);
#endif

	cmf = kmalloc(sizeof(struct csx_mem_file), GFP_KERNEL);
	if (NULL == cmf)
		return -ENOMEM;

	filp->private_data = cmf;
	atomic_set(&cmd->intr_count, 0);
	cmf->intr_count = atomic_read(&cmd->intr_count);
	/* open DMA interrupts*/
	writel(CSX_CTL_DMAINTR_MASK,
	       &cmd->maindata->ctl->regs[CSX_CTL_DMAINTRMASK_INDX]);
	MMIOWB();

	return 0;
}

/* Wait for a DMA interrupt */
static unsigned int
csx_mem_poll(struct file *filp, struct poll_table_struct *polltab)
{
	struct csx_mem_data *cmd;
	struct csx_mem_file *cmf;
	__s32 sample;

	cmd = container_of(filp->f_dentry->d_inode->i_cdev,
		struct csx_mem_data, cdev);
	cmf = filp->private_data;

	poll_wait(filp, &cmd->intr_queue, polltab);

	/* CSX poll implements a wait for interrupt */
	sample = atomic_read(&cmd->intr_count);
	if (cmf->intr_count != sample)
	{
		cmf->intr_count = sample;
		return (POLLIN | POLLRDNORM);
	}
	return 0;
}

/* Read device memory.
 * XXX: Ideally this code would know how to do DMAs.
 * XXX: Without DMAs, it should at least pin the target memory pages instead
 *	of blindly hoping they don't get remapped while the DMA is in flight.
 */
static ssize_t
csx_mem_read(struct file *filp, char __user * data, size_t size, loff_t * off)
{
	struct csx_mem_data *cmd;
	struct csx_mem_data *cmf;

	cmd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_mem_data, cdev);
	cmf = filp->private_data;

	/* Verify access doesn't fall outside memory area */
	if ((*off < 0) || (*off + size > cmd->len))
		return -EINVAL;

#if CS_NOMAPMEM
	/* Read from IO space */
	memcpy_fromio(data, &cmd->mem[*off], size);
#endif

	*off += size;
	return size;
}

/* Release per-file resources for memory aperture */
static int
csx_mem_release(struct inode *inode, struct file *filp)
{
	struct csx_mem_file *cmf;
	struct csx_mem_data *cmd;

	cmd = container_of(inode->i_cdev, struct csx_mem_data, cdev);
	// the dma mask is normally open for memory device
	/* close DMA interrupts*/
	writel(0, &cmd->maindata->ctl->regs[CSX_CTL_DMAINTRMASK_INDX]);
	MMIOWB();

	cmf = filp->private_data;
	kfree(cmf);

	return 0;
}

/* Write device memory
 * XXX: Ideally this code would know how to do DMAs.
 * XXX: Without DMAs, it should at least pin the source system memory pages
 *	instead of blindly hoping they don't get remapped while the DMA is in
 *	flight.
 */
static ssize_t
csx_mem_write(struct file *filp, const char __user * data, size_t size,
	      loff_t * off)
{
	struct csx_mem_data *cmd;
	struct csx_mem_file *cmf;

	cmd = container_of(filp->f_dentry->d_inode->i_cdev,
			   struct csx_mem_data, cdev);
	cmf = filp->private_data;

	/* Verify access doesn't fall outside memory area */
	if ((*off < 0) || (*off + size > cmd->len))
		return -EINVAL;

#if CS_NOMAPMEM
	/* Write from kernel space to IO space */
	/* XXX: On Altix this is dangerous, as you can run into the
	 * "PIO write storm" problem, which will panic the machine.
	 * Basically, PIO writes need to be slowed down.
	 */
	memcpy_toio(&cmd->mem[*off], (void *) data, size);

	/* Ensure writes arrive at hardware before proceeding */
	MMIOWB();
#endif

	*off += size;
	return size;
}

/******************
 * ioctl dispatch *
 ******************/
struct file_operations csx_mem_fops = {
	.owner = THIS_MODULE,
	.llseek = csx_mem_llseek,
	.mmap = csx_mem_mmap,
	.open = csx_mem_open,
	.poll = csx_mem_poll,
	.ioctl = csx_mem_ioctl,
	.read = csx_mem_read,
	.release = csx_mem_release,
	.write = csx_mem_write,
};

/********************
 * Class attributes *
 ********************/

/* Implements /sys/class/csx#c/dev */
static ssize_t
csx_show_ctl_dev(struct class_device *class_dev, char *buf)
{
	struct csx_device_data *cdd = class_get_devdata(class_dev);

	return print_dev_t(buf, cdd->ctl->dev);
}

/* Change this version whenver a released driver changes the format of
 * the /sys/class/csx/csx#c/status file.
 */
#define STATUS_VERSION "1"

/* Implements /sys/class/csx#c/status */
static ssize_t
csx_show_status(struct class_device *class_dev, char *buf)
{
	struct csx_device_data *cdd = class_get_devdata(class_dev);
	ssize_t len = 0;
	uint32_t fpga_version;

	/* Emit a status version, so that userland utilities can
	 * adapt to any changes in the output.
	 */
	len += snprintf(&buf[len], PAGE_SIZE - len,
			"Status version: %s\n", STATUS_VERSION);

	/* Driver build information */
	len += snprintf(&buf[len], PAGE_SIZE - len,
			"Driver build time: %s %s\n", __DATE__, __TIME__);

	/* Read FPGA version register */
	fpga_version = readl(&cdd->ctl->regs[CSX_CTL_FPGA_VERSION_INDX]);
	len += snprintf(&buf[len], PAGE_SIZE - len,
			"FPGA version: 0x%x\n", fpga_version);

	/* Additional items of interest (voltages, temps, etc.) follow
	 * this form:
	 *      value = readl(&cdd->ctl->regs[SOME_OFFSET]);
	 *      item = some_calculation(value);
	 *      len += snprintf(&buf[len], PAGE_SIZE-len,
	 *                      "Item name: %d\n", item);
	 *
	 * Ensure that the total buffer length doesn't exceed a single
	 * page.  The above code knows that the output is currently tiny,
	 * so it doesn't perform any such bounds checking.
	 */

	return len;
}

#define NULLATTR { .show = NULL, .store = NULL }

static struct class_device_attribute csx_ctl_attrs[] = {
	{
	 .attr = {.owner = THIS_MODULE,
		  .name = "dev",
		  .mode = 0444},
	 .show = csx_show_ctl_dev,
	 .store = NULL,
	 },
	{
	 .attr = {.owner = THIS_MODULE,
		  .name = "status",
		  .mode = 0444},
	 .show = csx_show_status,
	 .store = NULL,
	 },
	NULLATTR,
};

/* Implements /sys/class/csx#m/dev */
static ssize_t
csx_show_mem_dev(struct class_device *class_dev, char *buf)
{
	struct csx_device_data *cdd = class_get_devdata(class_dev);

	return print_dev_t(buf, cdd->mem->dev);
}

static struct class_device_attribute csx_mem_attrs[] = {
	{
	 .attr = {.owner = THIS_MODULE,
		  .name = "dev",
		  .mode = 0444},
	 .show = csx_show_mem_dev,
	 .store = NULL,
	 },
	NULLATTR
};

/*******************************
 * Device creation/destruction *
 *******************************/

/* Create device attribute files (e.g. "dev", "status") */
static int
csx_add_classdev_attrs(struct class_device *classdev,
		       struct class_device_attribute *attrs)
{
	struct class_device_attribute *curr;
	const char *attrname;
	int ret;

	for (curr = attrs; curr->show || curr->store; curr++) {
		ret = class_device_create_file(classdev, curr);
		if (ret) {
			attrname = curr->attr.name;
			goto out;
		}
	}

	return 0;
      out:
	while (curr != attrs) {
		class_device_remove_file(classdev, curr);
		curr--;
	}
	printk(KERN_WARNING "%s: Failed to add class device attribute "
	       "\"%s\"\n", __FUNCTION__, attrname);
	return ret;
}

/* Remove device attribute files */
static void
csx_remove_classdev_attrs(struct class_device *classdev,
			  struct class_device_attribute *attrs)
{
	struct class_device_attribute *curr;

	for (curr = attrs; curr->show || curr->store; curr++)
		class_device_remove_file(classdev, curr);
}

/*************************
 * PCI driver management *
 *************************/

static int
csx_setup_ctl(struct csx_device_data *cdd)
{
	int ret;
	char *errmsg;

	cdd->ctl = kmalloc(sizeof(struct csx_ctl_data), GFP_KERNEL);
	if (NULL == cdd->ctl)
		return -ENOMEM;
	memset(cdd->ctl, 0, sizeof(struct csx_ctl_data));
	cdd->ctl->maindata = cdd;

	/* Establish control register mappings */
	cdd->ctl->bar = pci_resource_start(cdd->pcidev, 0);
	if (0 == cdd->ctl->bar) {
		errmsg = "Unable to find control register resource";
		ret = -ENODEV;
		goto out_bar;
	}
	cdd->ctl->len = pci_resource_len(cdd->pcidev, 0);
	if (!request_mem_region(cdd->ctl->bar, cdd->ctl->len, "csx_ctl")) {
		errmsg = "Unable to obtain control register resource";
		ret = -ENODEV;
		goto out_bar;
	}
	cdd->ctl->regs = ioremap_nocache(cdd->ctl->bar, cdd->ctl->len);
	if (NULL == cdd->ctl->regs) {
		errmsg = "Unable to remap control register resource";
		ret = -ENODEV;
		goto out_remap;
	}

	/* Initialize interrupt count and wait queue */
	atomic_set(&cdd->ctl->intr_count, 0);
	init_waitqueue_head(&cdd->ctl->intr_queue);

	/* Create character special device */
	cdd->ctl->dev = cdd->dev_base + CSX_CTL_DEVNUM;
	cdd->ctl->cdev.owner = THIS_MODULE;
	kobject_set_name(&cdd->ctl->cdev.kobj, "csx%dc",
			 MINOR(cdd->ctl->dev) / CSX_NUM_DEVS);
	cdev_init(&cdd->ctl->cdev, &csx_ctl_fops);
	ret = cdev_add(&cdd->ctl->cdev, cdd->ctl->dev, 1);
	if (ret) {
		errmsg = "Failed to add control register cdev";
		goto out_cdev;
	}

	/* Add to device class */
	cdd->ctl->classdev.class = &csx_ctl_class;
	cdd->ctl->classdev.dev = &cdd->pcidev->dev;
	snprintf(cdd->ctl->classdev.class_id, BUS_ID_SIZE, "csx%dc",
		 MINOR(cdd->ctl->dev) / CSX_NUM_DEVS);
	class_set_devdata(&cdd->ctl->classdev, cdd);
	ret = class_device_register(&cdd->ctl->classdev);
	if (ret) {
		errmsg = "Failed to add control register class device";
		goto out_classdev;
	}

	/* Create attribute files */
	ret = csx_add_classdev_attrs(&cdd->ctl->classdev, csx_ctl_attrs);
	if (ret) {
		errmsg = "Failed to add control register class device "
			"attributes";
		goto out_attr;
	}

	/* Mask PCI interrupts.  Mask is opened before poll by user code. */
	writel(0, &cdd->ctl->regs[CSX_CTL_PCIINTRMASK_INDX]);
	MMIOWB();

	return 0;

	/* Before the class device is registered, we need
	 * to perform error teardown ourself.
	 */
out_classdev:
	cdev_del(&cdd->ctl->cdev);
out_cdev:
	iounmap(cdd->ctl->regs);
out_remap:
	release_mem_region(cdd->ctl->bar, cdd->ctl->len);
out_bar:
	kfree(cdd->ctl);
	goto out;

	/* After the class device is registered, the release
	 * method performs error teardown.
	 */
out_attr:
	class_device_unregister(&cdd->ctl->classdev);
out:
	printk(KERN_WARNING "%s: %s for pci_dev 0x%p\n",
	       __FUNCTION__, errmsg, cdd->pcidev);
	return ret;
}

static void
csx_teardown_ctl(struct csx_device_data *cdd)
{
	/* Remove control register device */
	csx_remove_classdev_attrs(&cdd->ctl->classdev, csx_ctl_attrs);
	class_device_unregister(&cdd->ctl->classdev);
}

static int
csx_setup_mem(struct csx_device_data *cdd)
{
	int ret;
	char *errmsg;
	void *desc_coherent_buffer;
	dma_addr_t desc_bus_address;
#if CSX_BOUNCE
	void  *data_coherent_buffer;
	dma_addr_t data_bus_address;
	unsigned target_size;
#endif

	cdd->mem = kmalloc(sizeof(struct csx_mem_data), GFP_KERNEL);
	if (NULL == cdd->mem) {
		errmsg = "Failed to allocate memory device data structure";
		ret = -ENOMEM;
		goto out;
	}
	memset(cdd->mem, 0, sizeof(struct csx_mem_data));
	cdd->mem->maindata = cdd;

	/* Set DMA address mask */
	dma_set_mask(&cdd->pcidev->dev, ~0ULL);
	if( pci_set_consistent_dma_mask(cdd->pcidev, ~0ULL) )
    {
        errmsg = "Failed to set consistent DMA mask";
		ret = -ENOMEM;
		goto out_desc;
    }

	/* Mask DMA interrupts.  Mask is opened once a client connects. */
	writel(0, &cdd->ctl->regs[CSX_CTL_DMAINTRMASK_INDX]);
	MMIOWB();

	/* Obtain coherent mapping for the dma descriptor buffer.
	 * this will be remapped to user space in the dma mmap function
	 * These coherent mappings remain for the life of the driver. 
	 */

    /* Allocating two buffers so that we may prepare the second
     * buffer whilst DMA'ing the first
     */
	desc_coherent_buffer = dma_alloc_coherent(&cdd->pcidev->dev,
						  MAX_DESCRIPTOR_BUFFER_SIZE * 2,
						  &desc_bus_address,
						  GFP_KERNEL);

	if (!desc_coherent_buffer) {
		errmsg = "Failed to allocate descriptor coherent buffer";
		ret = -ENOMEM;
		goto out_desc;
	}
	cdd->mem->region.desc_bus_address[0] = desc_bus_address;
	cdd->mem->region.desc_bus_address[1] = desc_bus_address + MAX_DESCRIPTOR_BUFFER_SIZE;
	cdd->mem->region.desc_kaddr = desc_coherent_buffer;
	cdd->mem->region.desc_buf_size = MAX_DESCRIPTOR_BUFFER_SIZE;

#if CSX_BOUNCE
	/* Obtain a single coherent DMA mapping if DMA bounce buffer needed. */
	for (target_size = CSX_BOUNCE_SIZE;
	     target_size >= CSX_SMALLEST_BOUNCE_SIZE; target_size /= 2) {
		data_coherent_buffer =
			dma_alloc_coherent(&cdd->pcidev->dev, target_size,
					   &data_bus_address, GFP_KERNEL|__GFP_NOWARN);
		if (data_coherent_buffer)
			break;
	}
	if (!data_coherent_buffer) {
		errmsg = "Failed to allocate data coherent buffer";
		ret = -ENOMEM;
		goto out_data;
	}
	cdd->mem->region.data_bus_address = data_bus_address;
	cdd->mem->region.data_kaddr = data_coherent_buffer;
	cdd->mem->region.data_buf_size = target_size;	/* User side will limit requests to this */
#if CSX_DEBUG
	printk(KERN_DEBUG "%s: allocated data coherent buffer bus %llx, kernel %p, size %x\n",
	       __FUNCTION__, data_bus_address, data_coherent_buffer, target_size);
#endif
#endif

	/* Establish memory mappings */
	cdd->mem->bar = pci_resource_start(cdd->pcidev, 2);
	if (0 == cdd->mem->bar) {
		errmsg = "Unable to find memory resource";
		ret = -ENODEV;
		goto out_bar;
	}
	cdd->mem->len = pci_resource_len(cdd->pcidev, 2);
	if (!request_mem_region(cdd->mem->bar, cdd->mem->len, "csx_mem")) {
		errmsg = "Unable to obtain memory resource";
		ret = -ENODEV;
		goto out_bar;
	}
#if CS_NOMAPMEM
	cdd->mem->mem = ioremap(cdd->mem->bar, cdd->mem->len);
	if (NULL == cdd->mem->mem) {
		errmsg = "Unable to remap memory resource";
		ret = -ENODEV;
		goto out_remap;
	}
#else
	cdd->mem->mem = 0;
#endif

	/* Initialize interrupt count and wait queue */
	atomic_set(&cdd->mem->intr_count, 0);
	init_waitqueue_head(&cdd->mem->intr_queue);

	/* Create character special device */
	cdd->mem->dev = cdd->dev_base + CSX_MEM_DEVNUM;
	cdd->mem->cdev.owner = THIS_MODULE;
	kobject_set_name(&cdd->mem->cdev.kobj, "csx%dm",
			 MINOR(cdd->mem->dev) / CSX_NUM_DEVS);
	cdev_init(&cdd->mem->cdev, &csx_mem_fops);
	ret = cdev_add(&cdd->mem->cdev, cdd->mem->dev, 1);
	if (ret) {
		errmsg = "Failed to add memory cdev";
		goto out_cdev;
	}

	/* Add to device class */
	cdd->mem->classdev.class = &csx_mem_class;
	cdd->mem->classdev.dev = &cdd->pcidev->dev;
	snprintf(cdd->mem->classdev.class_id, BUS_ID_SIZE, "csx%dm",
		 MINOR(cdd->mem->dev) / CSX_NUM_DEVS);
	class_set_devdata(&cdd->mem->classdev, cdd);
	ret = class_device_register(&cdd->mem->classdev);
	if (ret) {
		errmsg = "Failed to add memory class device";
		goto out_classdev;
	}

	/* Create attribute files */
	ret = csx_add_classdev_attrs(&cdd->mem->classdev, csx_mem_attrs);
	if (ret) {
		errmsg = "Failed to add memory class device attributes";
		goto out_attr;
	}

	return 0;

	/* Before the class device is registered, we need
	 * to perform error teardown ourself.
	 */
out_classdev:
	cdev_del(&cdd->mem->cdev);
out_cdev:
#if CS_NOMAPMEM
	iounmap(cdd->mem->mem);
out_remap:
#endif
	release_mem_region(cdd->mem->bar, cdd->mem->len);
out_bar:
#if CSX_BOUNCE
	dma_free_coherent(&cdd->pcidev->dev, cdd->mem->region.data_buf_size,
			  data_coherent_buffer, data_bus_address);
out_data:
#endif
	dma_free_coherent(&cdd->pcidev->dev, MAX_DESCRIPTOR_BUFFER_SIZE * 2,
			  desc_coherent_buffer, desc_bus_address);
out_desc:
	kfree(cdd->mem);
	goto out;

	/* After the class device is registered, the release
	 * method performs error teardown.
	 */
out_attr:
	class_device_unregister(&cdd->mem->classdev);
out:
	printk(KERN_WARNING "%s: %s for pci_dev 0x%p\n",
	       __FUNCTION__, errmsg, cdd->pcidev);
	return ret;
}

static void
csx_teardown_mem(struct csx_device_data *cdd)
{
	/* Remove memory device */
	csx_remove_classdev_attrs(&cdd->mem->classdev, csx_mem_attrs);
	class_device_unregister(&cdd->mem->classdev);
}

/* Sets up driver resources upon discovery of a new CSX device */
static int
csx_probe(struct pci_dev *pdev, const struct pci_device_id *pci_id)
{
	int ret;
	char *errmsg;
	struct csx_device_data *cdd;

    /* Enable device and turn on bus mastering */
	ret = pci_enable_device(pdev);
	if (ret) {
		errmsg = "Failed to enable device";
		goto out_enable;
	}
	pci_set_master(pdev);

	/* Create per-CSX data */
	cdd = kmalloc(sizeof(struct csx_device_data), GFP_KERNEL);
	if (NULL == cdd) {
		errmsg = "Failed to allocate device data";
		ret = -ENOMEM;
		goto out_kmalloc;
	}
#if CSX_DEBUG
        printk(KERN_DEBUG "%s: created cdd %p\n", __FUNCTION__, cdd);
#endif
	pci_set_drvdata(pdev, cdd);
	cdd->pcidev = pdev;

	/* Reserve minor device numbers for this board */
	spin_lock(&nextdev_lock);
	cdd->dev_base = nextdev;
	nextdev += CSX_NUM_DEVS;
	spin_unlock(&nextdev_lock);

	/* Set up individual devices */
	if (0 != (ret = csx_setup_ctl(cdd))) {
		errmsg = "Failed to setup control device";
		goto out_ctl;
	}
	if (0 != (ret = csx_setup_mem(cdd))) {
		errmsg = "Failed to setup memory device";
		goto out_mem;
	}

	/* Register interrupt handler */
	ret = request_irq(pdev->irq, csx_intr_handler, SA_SHIRQ, "csx", cdd);
	if (ret) {
		errmsg = "Failed to register interrupt handler";
		goto out_handler;
	}
	return 0;

out_handler:
	csx_teardown_mem(cdd);
out_mem:
	csx_teardown_ctl(cdd);
out_ctl:
	kfree(cdd);
    /* Try to keep successful devices in numerical sequence */
	nextdev -= CSX_NUM_DEVS;
out_kmalloc:
	pci_disable_device(pdev);
out_enable:
	printk(KERN_WARNING "%s: %s for pci_dev 0x%p\n",
	       __FUNCTION__, errmsg, pdev);
	return ret;
}

/* Removes driver resources upon removal of an existing CSX device */
static void
csx_remove(struct pci_dev *pdev)
{
	struct csx_device_data *cdd;

	cdd = pci_get_drvdata(pdev);

	free_irq(cdd->pcidev->irq, cdd);
	csx_teardown_mem(cdd);
	csx_teardown_ctl(cdd);
	pci_disable_device(pdev);
	kfree(cdd);
}

#define PCI_VENDOR_ID_CLEARSPEED 0x1942
#define PCI_DEVICE_ID_CLEARSPEED_CSX 0xe511

static struct pci_device_id csx_id_table[] = {
	{PCI_DEVICE(PCI_VENDOR_ID_CLEARSPEED, PCI_DEVICE_ID_CLEARSPEED_CSX)},
	{0}
};

MODULE_DEVICE_TABLE(pci, csx_id_table);

static struct pci_driver __devinitdata csx_driver = {
	.name = "CSX",
	.id_table = csx_id_table,
	.probe = csx_probe,
	.remove = csx_remove,
};

/*********************
 * Module management *
 *********************/

/* Get dynamic device numbers and register device class */
static int __devinit
csx_init(void)
{
	char *errmsg;
	int ret;

#if CSX_DEBUG
    printk(KERN_DEBUG "Initialising OSD Driver.\n" );
#endif

	/* Get block of dynamic device numbers */
	ret = alloc_chrdev_region(&firstdev, 0, 256, "csx");
	if (ret) {
		errmsg = "Failed to allocate device number block";
		goto out_dev;
	}
	nextdev = firstdev;
	nextdev_lock = SPIN_LOCK_UNLOCKED;

	/* Class to allow userland to find the device numbers and
	 * automatically create /dev entries.
	 */
	ret = class_register(&csx_ctl_class);
	if (ret) {
		errmsg = "Failed to create control device class";
		goto out_ctl_class;
	}
	ret = class_register(&csx_mem_class);
	if (ret) {
		errmsg = "Failed to create memory device class";
		goto out_mem_class;
	}

	/* Register device driver */
	pci_register_driver(&csx_driver);

	return 0;

out_mem_class:
	class_unregister(&csx_ctl_class);
out_ctl_class:
	unregister_chrdev_region(firstdev, 256);
out_dev:
	printk(KERN_WARNING "%s: %s (%d)\n", __FUNCTION__, errmsg, ret);
	return ret;
}

/* Unregister device class and free dynamic device numbers. */
static void __devexit
csx_exit(void)
{
	/* Unregister device driver */
	pci_unregister_driver(&csx_driver);

	/* Remove device classes */
	class_unregister(&csx_mem_class);
	class_unregister(&csx_ctl_class);

	/* Release device number block */
	unregister_chrdev_region(firstdev, 256);

#if CSX_DEBUG
    printk(KERN_DEBUG "Exiting OSD Driver.\n" );
#endif
}

module_init(csx_init);
module_exit(csx_exit);

MODULE_AUTHOR("Silicon Graphics, Inc. / ClearSpeed Technology Plc");
MODULE_DESCRIPTION("ClearSpeed Advance CSX600 device driver");
MODULE_LICENSE("GPL");
