//
//
//      Copyright (c) ClearSpeed Technology plc 2000,2002,2006
//      All rights reserved.
//
//      Copyright (C) 2006 Silicon Graphics, Inc.  All Rights Reserved.
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk. Permission to use or copy this
// software for any purpose is hereby granted without fee, provided the
// above notices are retained on all copies. Permission to modify the code
// and to distribute modified code is granted, provided the above notices
// are retained, and a notice that the code was modified is included with
// the above copyright notice.
//
//
#ifndef LIBKERNELUSER_H
#define LIBKERNELUSER_H


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Common definitions
//
///////////////////////////////////////////////////////////////////////////////////////////////////
#define MAX_SINGLE_DMA_TRANSFER                 (4 * 1024 * 1024)
#define HALFBRIDGE_DESCRIPTOR_SIZE              32


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  These definitons then follow
//  Maximum number of scatter gather entries, also maximum number of descriptor entires
//
///////////////////////////////////////////////////////////////////////////////////////////////////
#define MAX_SG_SIZE                             ((MAX_SINGLE_DMA_TRANSFER/PAGE_SIZE) + 2)
#define MAX_DESCRIPTOR_BUFFER_SIZE              ((MAX_SG_SIZE) * HALFBRIDGE_DESCRIPTOR_SIZE)


// Sizes for coherent bounce buffers (if used)
#define CSX_BOUNCE_SIZE                         MAX_SINGLE_DMA_TRANSFER
#define CSX_SMALLEST_BOUNCE_SIZE                (CSX_BOUNCE_SIZE>>3)


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  ioctl definitions
//
///////////////////////////////////////////////////////////////////////////////////////////////////
#define CSX_IOCTL_MAGIC                         0xC5   // aka clearspeed

#define CSX_LOCK_DATA_BUFFER         _IOWR(CSX_IOCTL_MAGIC, 0x90, struct ioctl_param_buffer_lock)
#define CSX_UNLOCK_DATA_BUFFER       _IOWR(CSX_IOCTL_MAGIC, 0x91, struct ioctl_param_buffer_unlock)
#define CSX_GET_COHERENT_BUFFER_INFO _IOR(CSX_IOCTL_MAGIC,  0x92, struct \
                                                ioctl_param_get_coherent_map_info)
#define CSX_COPY_TO_DESC_BUFFER      _IOR(CSX_IOCTL_MAGIC,  0x93, struct \
                                                ioctl_param_copy_to_coherent_descriptor_buffer)
#define CSX_COPY_DATA_BUFFER         _IOR(CSX_IOCTL_MAGIC,  0x94, struct \
                                                ioctl_param_copy_data_buffer)
#define CSX_FIRE_DMA                 _IOR(CSX_IOCTL_MAGIC,  0x95, struct ioctl_param_buffer_fire)
#define CSX_COPY_AND_FIRE            _IOR(CSX_IOCTL_MAGIC,  0x96, struct ioctl_param_copy_and_fire)
#define CSX_GET_PAGE_SIZE            _IOR(CSX_IOCTL_MAGIC,  0x97, struct ioctl_param_get_page_size)


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Page descriptor, typedefs and flags
//
///////////////////////////////////////////////////////////////////////////////////////////////////
typedef unsigned long long bus_address_t;
typedef int status;

struct dma_desc {
        void            *virt_address;
        bus_address_t   bus_address;
        unsigned        offset;
        unsigned        length;
};

#define DMA_FLAG_TO_DEVICE                      0x1
#define DMA_FLAG_FROM_DEVICE                    0x2
#define DMA_FLAG_BIDIRECTONAL                   (DMA_FLAG_TO_DEVICE|DMA_FLAG_FROM_DEVICE)
#define DMA_FLAG_CONTROL                        0x4


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// The ioctl structures
//
///////////////////////////////////////////////////////////////////////////////////////////////////
struct ioctl_param_buffer_lock 
{
    int buffer;
    void * user_address;
    unsigned user_size;
    int desc_count;
    struct dma_desc * desc_array;
    unsigned flags;
};

struct ioctl_param_buffer_unlock
{
    int buffer;
    int desc_count;
    struct dma_desc *desc_array;
    unsigned flags;
};

struct ioctl_param_buffer_fire
{
    int buffer;
    bus_address_t desc_address;
    unsigned flags;
};

struct ioctl_param_copy_to_coherent_descriptor_buffer
{
    int buffer;
    void * user_descriptor_buffer;
    unsigned size;
    bus_address_t desc_address;
};

struct ioctl_param_copy_data_buffer
{
    void * user_data_buffer;
    unsigned size;
    unsigned flags;
};

struct ioctl_param_get_coherent_map_info
{
    bus_address_t descriptor_chain1;
    bus_address_t descriptor_chain2;
    bus_address_t bounce_buffer;
    unsigned int bounce_buffer_size;
};

struct ioctl_param_copy_and_fire
{
    int buffer;
    void * user_descriptor_buffer;
    unsigned size;
    bus_address_t desc_address;
    unsigned flags;
};

struct ioctl_param_get_page_size
{
    unsigned int page_size_in_bytes;
};


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// lock_buffer_for_dma: Locks a user-supplied buffer into memory for DMA transfers, returning bus
//                      address mappings.
// Input
//      user_address    Virtual address of beginning of buffer to be transferred.
//
//      user_size       Length of buffer to be transferred.
//
//      desc_count      Number of dma_desc entries supplied by caller via desc_array
//
//      desc_array      Array of desc_desc descriptors to be filled in
//
//      flags           Set of flags controlling allocation/mapping
//                      DMA_FLAG_TO_DEVICE      DMA direction will be to CSX600 from main memory
//                      DMA_FLAG_FROM_DEVICE    DMA direction will be from CSX600 to main memory
//                      DMA_FLAG_BIDIRECTIONAL  DMA direction will be both to and from CSX600
//                                              to/from main memory
//                      DMA_FLAG_CONTROL        Mapping is for DMA control descriptor chain (as
//                                              opposed to DMA contents themselves)
// Output
//      desc_count      If input desc_count was sufficient to cover entire transfer, and enough I/O
//                      translation resources are present, on output desc_count contains the actual
//                      number of entries used in page_array. The return value of the function will
//                      be 0.
//
//                      If input desc_count was sufficient to cover entire transfer, but not enough
//                      I/O translation resources were available to cover the entire transfer, on
//                      output desc_count contains the actual number of entries used in desc_array,
//                      which will only cover a portion of the transfer.  The return value of the
//                      function will be EAGAIN.
//
//                      If input desc_count was insufficient to cover entire transfer, on output
//                      desc_count contains the number of entries necessary to cover the entire
//                      transfer, and no descriptors in desc_array are valid. The return value of
//                      the function will be E2BIG.
//
//      desc_array      In the case of complete or partial transfer setup (first two cases
//                      described under output value of desc_count), the first desc_count entries
//                      in desc_array will be filled in.
//
//                      Each entry consists of the following:
//                              virt_address: First virtual address covered by this descriptor for
//                                            which an I/O mapping has been established.
//                              bus_address:  First PCI bus address covered by this descriptor for
//                                            which an I/O mapping has been established.
//                              offset:       Offset into {virt,phys}_address of first byte to be
//                                            transferred.
//                              length:       Number of bytes to transfer in this descriptor.
//
//                      In the case not enough entries were supplied as input, the output values of
//                      desc_array are undefined.
//
// Return
//      See descriptions under output of desc_count.
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status lock_buffer_for_dma(int fd, void *user_address, unsigned user_size, int *desc_count,
                           struct dma_desc *desc_array, unsigned flags);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// unlock_buffer_for_dma: Unlocks a previously locked DMA buffer from memory, tearing down bus
//                         address mappings.
// Input
//      desc_count: The output desc_count from the preceding call to lock_buffer_for_dma().
//      desc_array: The output desc_array from the preceding call to lock_buffer_for_dma().
//      flags:      The same set of flags as was passed to the preceding lock_buffer_for_dma.
//
// Output
//      None.
//
// Return
//      Returns 0 on success, nonzero otherwise.
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status unlock_buffer_for_dma(int fd, int desc_count, struct dma_desc *desc_array, unsigned flags);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// fire_dma: Fire a previously prepared dma operation. 
//
// Input
//    bus_chain_start: Address of descriptor chain as a bus (pci/device) address.
//    flags:           The same set of flags as was passed to the preceding lock_buffer_for_dma.
//
// Output
//    None
//
// Return
//    Returns 0 on success, nonzero otherwise
//
// Notes
//    This ioctl fires the dma by setting up the first descriptor chain address
//    and firing the dma. So, different stratigies can be used for dma. the only
//    requirement is that a sensible chain address is used. The ioctl will check that
//    the address is either from the coherent chain or from the locked and pinned chain.
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status fire_dma(int fd, bus_address_t desc_chain_start, unsigned flags);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// get_coherent_map_info: Get bus address info for coherent mappings
//                        Returns the location of the descriptor_chain buffer and the location and
//                        size of the coherent bounce_buffer
// Input:  None
//
// Output:
//         descriptor_chain1: bus address for descriptor chain1
//         descriptor_chain2: bus address for descriptor chain2
//         bounce_buffer: bus address for bounce buffer
//         bounce_buffer_size: size of the bounce buffer.
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status get_coherent_map_info(int fd, bus_address_t * const descriptor_chain1,
                                bus_address_t * const descriptor_chain2,
                                bus_address_t * const bounce_buffer,
                                unsigned int * bounce_buffer_size);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// copy_to_coherent_descriptor_buffer: Copy the descriptor buffer prepared by the user libary into
//                                     the dma coherent buffer used for the descriptor
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status copy_to_coherent_descriptor_buffer(int fd, const void * const user_descriptor_buffer,
                                unsigned int size);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// copy_data_buffer: Depending on the flags, copy either to or from the coherent bounce buffer.
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status copy_data_buffer(int fd, void * const user_buffer, unsigned sizeo, unsigned flags);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// copy_and_fire: Equivalent to copy_to_coherent_descriptor_buffer followed by fire_dma. Uses one
//                kernel transition for increased performance
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status copy_and_fire(int fd, const void * const user_descriptor_buffer, unsigned int size,
                                bus_address_t desc_chain_start, unsigned flags);


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// get_page_size: Extract the machine kernel page size
//
///////////////////////////////////////////////////////////////////////////////////////////////////
status get_page_size(int fd, unsigned int *p_page_size_in_bytes);


#endif // LIBKERNELUSER_H
