/*
 *
 *     Copyright (c) 2006, ClearSpeed Technology plc. All rights reserved.
 *
 * This material is provided "as is", with absolutely no warranty expressed or
 * implied. Any use is at your own risk. Permission to use or copy this software
 * for any purpose is hereby granted without fee, provided the above notices are
 * retained on all copies. Permission to modify the code and to distribute
 * modified code is granted, provided the above notices are retained, and a
 * notice that the code was modified is included with the above copyright notice.
 *
 */


/*!
  \file   cs_dft_demo_1d_z_user.c
  \brief  Clearspeed DFT Library Test

  \date   30-January-2006

  This test demonstrates the creation of double precision 1d complex user plan
  and it's execution. It cycles through sizes 128 to 8192 in powers of 2,
  creating random data and then allocating a memory descriptor to the data.
  It then creates a 1d double precision complex user plan. This user plan
  does a forward fft and a backward one, based on the parameters of the user
  plan. The actual board based user routine is built under the top level
  user_csx directory and is loaded at the start by setting the environment
  variable CS_FFT_CSX to be the base name of the csx files (i.e. without the
  suffix of _0.csx, _1_debug.csx etc) before running the program. There should
  exist a csx file for each chip.

  It executes the user plan, such that the resulting output should be
  almost identical to the original input.

*/

#include <math.h>
#include <stdlib.h>

#include "csdft.h"
#include "cs_complex.h"

struct csdft_user_data_defn
{
    unsigned int flags;
    unsigned int size_x;
    unsigned int size_y;
    unsigned int unused;
    double       scale;
    unsigned int dimensions;
    unsigned int number_of_ffts;
} sdft_user_data;

CSDftUserFunc user_download_control_func = NULL;
CSDftUserFunc user_func = NULL;

/*
    Routines to help compare input and golden image
*/
unsigned int compare_value(double val, double gold_val)
{
    double diff;

    if ( fabs(gold_val) < 0.001)
    {
        diff = fabs(val-gold_val);
    } else
    {
        diff = fabs((val-gold_val)/gold_val);
    }
    if (diff > 0.05)
    {
        return 0;
    } else
    {
        return 1;
    }
}


unsigned int compareFft(DoubleComplex *result, DoubleComplex *gold, unsigned int w, unsigned int h)
{
    unsigned int r,c;

    for (r = 0; r < h; r++)
        for (c = 0; c < w; c++)
        {
            if (!compare_value(CS_crealf(result[r*w+c]), CS_crealf(gold[r*w+c]))||
                !compare_value(CS_cimagf(result[r*w+c]), CS_cimagf(gold[r*w+c])))
            {
                printf("Failed fft comparison locn, column %d, row %d\n", c, r);
                return 0;
            }
        }
    return 1;
}

/*
    This routine creates 2 1d double precision complex plans
    One does a forward fft, the other a backward fft.
    The resulting output should be the original input
*/
unsigned int cs_dft_user(unsigned int n)
{
    unsigned int i = 0, j = 0;
    unsigned int num_ffts = 100;
    unsigned int passed = 1;
    double v,vi;
    DoubleComplex *in = NULL, *out = NULL;
    struct csdft_user_data_defn *desc = NULL;
    CSDftMemoryDescriptor md_in;
    CSDftMemoryDescriptor md_out;
    CSDftMemoryDescriptor init_dat;
    CSDftStatus           status;
    CSDftPlan             plan;


    /* malloc local space for ffts */
    in     = malloc(sizeof(DoubleComplex) * n *num_ffts);
    out    = malloc(sizeof(DoubleComplex) * n *num_ffts);
   
    /* Generate some random input */
    srand(1);/* Error check */
    if (CSDFT_get_status() != CSDFT_NO_ERROR)
    {
    
       /* An error has occurred. Tell user the problem: */
       printf("%s\n", CSDFT_return_error_message(CSDFT_get_status()));
 
       /* Fail the test */
       passed = 0;
       return passed;
    
    }

    for (j = 0; j < num_ffts; j++)
    {
        for (i=0; i<n ; i++)
        {
            do {
                v  = (1.0*rand())/RAND_MAX;
                vi = (1.0*rand())/RAND_MAX;
            } while (v < 0.02 || vi < 0.02);

            CS_complex_assign(in[j*n+i], v, vi);
        }
    }
    /* Allocate pointer to structure describing DFT */
    desc =&sdft_user_data;

    /* Create the host memory descriptors that point to the pre-allocated in and out arrays */
    md_in  = CSDFT_system_to_memory_descriptor((void *)in,  (sizeof(DoubleComplex) * n) * num_ffts);
    md_out = CSDFT_system_to_memory_descriptor((void *)out, (sizeof(DoubleComplex) * n) * num_ffts);
    
    /* Create host memory descriptor that points to pre-allocated stucture defining DFT */
    init_dat = CSDFT_system_to_memory_descriptor((void *)desc, sizeof(struct csdft_user_data_defn));
/* Error check */
    if (CSDFT_get_status() != CSDFT_NO_ERROR)
    {
    
       /* An error has occurred when processing kernel. Tell user the problem: */
       printf("%s\n", CSDFT_return_error_message(CSDFT_get_status()));
 
       /* Fail the test */
       passed = 0;
       return passed;
    
    }

    /*
     Create the plan that describes the DFTs, in this case calling a user
     routine which does a double precision forward, then a backward DFT.

     The scale factor is only applied to the backward DFT.
    */
    sdft_user_data.flags      = CSDFT_FORWARD|CSDFT_OPTIMAL_ORDER|CSDFT_COMPLEX|CSDFT_DOUBLE;
    sdft_user_data.size_x     = n;
    sdft_user_data.size_y     = n;
    sdft_user_data.scale      = 1.0/n;
    sdft_user_data.dimensions = 1;
    sdft_user_data.number_of_ffts = num_ffts;
   
    /* Create the handle to the user function */
    user_func =   CSDFT_create_user_func("user_forward_backward_dft", sizeof(struct csdft_user_data_defn), sizeof(DoubleComplex)*n* num_ffts, sizeof(DoubleComplex) * n * num_ffts, 1);

    /* 
      Execute the function.  The last parameter defines how data is distributed to the processors on the board.
      CSDFT_BOARD_SEQUENCE and CSDFT_PROCESSOR_SEQUENCE means that data packets are sent to alternate boards and processors
      in sequence.  More details available in the documentation.
    */  
    status = CSDFT_execute_user_func(user_func, init_dat, md_in, md_out, CSDFT_BOARD_SEQUENCE | CSDFT_PROCESSOR_SEQUENCE);
    
    /* Error check */
    if ((CSDFT_get_status() != CSDFT_NO_ERROR)&&(status != CSDFT_NO_ERROR))
    {
    
       /* An error has occurred when processing kernel. Tell user the problem: */
       printf("%s\n", CSDFT_return_error_message(CSDFT_get_status()));
 
       /* Fail the test */
       passed = 0;
       return passed;
    
    }


    /* Check that the input and output are almost the same */
    for (j = 0; j < num_ffts; j++)
    {
        if (!compareFft(&out[j*n], &in[j*n], n, 1))
        {
            printf("DFT %d failed for size %d\n", j+1, n);
            passed = 0;
            break;
        }
    }


    /* Free the memory */
    free(out);
    free(in);

    /* Free the memory descriptors */
    CSDFT_free(md_out);
    CSDFT_free(md_in);
    CSDFT_free(init_dat);



    return passed;
}

/*
    Main
*/
int main(int argc, char *argv[])
{
    unsigned int s;
    unsigned int passed = 1;

    for (s = 128; s <= 8192; s = s*2)
    {
        if (!cs_dft_user(s))
        {
            printf("Failed 1D double precision complex user DFT of size %d\n\n", s);
            passed  = 0;
        }
    }

    printf("\n**** 1D complex double precision user test: ");

    if (passed)
    {
        printf("passed\n");
    } else
    {
        printf("failed\n");
    }
    printf("\n");

    return 0;
}

