/*
 *
 *     Copyright (c) 2006, ClearSpeed Technology plc. All rights reserved.
 *
 * This material is provided "as is", with absolutely no warranty expressed or
 * implied. Any use is at your own risk. Permission to use or copy this software
 * for any purpose is hereby granted without fee, provided the above notices are
 * retained on all copies. Permission to modify the code and to distribute
 * modified code is granted, provided the above notices are retained, and a
 * notice that the code was modified is included with the above copyright notice.
 *
 */


/*!
  \file   cs_dft_demo_1d_c_convolution.c
  \brief  Clearspeed DFT Library Test

  \date   15-March-2006

  This test demonstrates the creation of two single precision 1d complex user
  plans which between them create a convolution kernel and perform the
  convolution against another input array. It cycles through sizes 128 to 8192
  in powers of 2,

  The actual board based user routine is built under the top level
  user_csx directory and is loaded at the start by setting the environment
  variable CS_FFT_CSX to be the base name of the csx files (i.e. without the
  suffix of _0.csx, _1_debug.csx etc) before running the program. There should
  exist a csx file for each chip.

  It executes the user plan, such that the resulting output should be
  almost identical to the original input.

*/

#include <math.h>
#include <stdlib.h>

#include "csdft.h"
#include "cs_complex.h"

struct csdft_user_data_defn
{
    unsigned int flags;
    unsigned int size_x;
    unsigned int size_y;
    unsigned int unused;
    double       scale;
    unsigned int dimensions;
    unsigned int number_of_ffts;
} csdft_user_data;

CSDftUserFunc user_download_control_func = NULL;
CSDftUserFunc user_func = NULL, conv_func= NULL;

/*
    Routines to help compare input and golden image
*/
unsigned int compare_value(float val, float gold_val)
{
    float diff;

    if ( fabsf(gold_val) < 0.001)
    {
        diff = fabsf(val-gold_val);
    } else
    {
        diff = fabsf((val-gold_val)/gold_val);
    }
    if (diff > 0.05)
    {
        return 0;
    } else
    {
        return 1;
    }
}


unsigned int compareFft(FloatComplex *result, FloatComplex *gold, unsigned int w, unsigned int h)
{
    unsigned int r,c;

    for (r = 0; r < h; r++)
        for (c = 0; c < w; c++)
        {
            if (!compare_value(CS_crealf(result[r*w+c]), CS_crealf(gold[r*w+c]))||
                !compare_value(CS_cimagf(result[r*w+c]), CS_cimagf(gold[r*w+c])))
            {
                printf("Result: (%f, %f)\nGolden: (%f, %f)\n", CS_crealf(result[r*w+c]), CS_cimagf(result[r*w+c]), CS_crealf(gold[r*w+c]), CS_cimagf(gold[r*w+c]));

                printf("Failed fft comparison locn, column %d, row %d\n", c, r);
                return 0;
            }
        }
    return 1;
}

/*
    Single precision complex 1d user routine for size n
*/
unsigned int cs_dft_user_convolution(unsigned int n)
{
    unsigned int i = 0, j = 0;
    unsigned int num_ffts = 1;
    unsigned int passed   = 1;

    float v,vi;

    FloatComplex *in = NULL, *out = NULL, *kernel = NULL;

    CSDftMemoryDescriptor md_in;
    CSDftMemoryDescriptor md_out;
    CSDftMemoryDescriptor md_null;
    CSDftMemoryDescriptor md_kernel;
    CSDftMemoryDescriptor md_init;

    CSDftStatus           status;

    CSDftPlan             plan_kernel;
    CSDftPlan             plan_convolution;


    /* malloc local space for ffts */
    in     = malloc(sizeof(FloatComplex) * n * num_ffts);
    out    = malloc(sizeof(FloatComplex) * n * num_ffts);
    kernel = calloc(sizeof(FloatComplex), n );


    /* Generate some random input for input*/
    srand(1);

    for (j = 0; j < num_ffts; j++)
    {
        for (i=0; i<n ; i++)
        {
            do {
                v  = (1.0*rand())/RAND_MAX;
                vi = (1.0*rand())/RAND_MAX;
            } while (v < 0.5 || vi < 0.5);

            CS_complex_assign(in[j*n+i], v, vi);
        }
    }

    /* Generates identity matrix after DFT of kernel */
    CS_complex_assign(kernel[0], 1.0, 0.0);


    /* Create the host memory descriptors that point to the pre-allocated in and out arrays */
    md_in     = CSDFT_system_to_memory_descriptor((void *)in,     (sizeof(FloatComplex) * n)* num_ffts);
    md_out    = CSDFT_system_to_memory_descriptor((void *)out,    (sizeof(FloatComplex) * n)* num_ffts);
    md_kernel = CSDFT_system_to_memory_descriptor((void *)kernel, (sizeof(FloatComplex) * n));
    md_init = CSDFT_system_to_memory_descriptor((void *)&csdft_user_data, sizeof(struct csdft_user_data_defn));
    md_null = CSDFT_get_null_descriptor() ;

    /* Describe the kernel DFT transformation */
    csdft_user_data.flags      = CSDFT_FORWARD|CSDFT_OPTIMAL_ORDER|CSDFT_COMPLEX|CSDFT_SINGLE;
    csdft_user_data.size_x     = n;
    csdft_user_data.scale      = 1.0;
    csdft_user_data.dimensions = 1;
    csdft_user_data.number_of_ffts = num_ffts;

    /* Create the function that will perform the kernel DFT */
    user_func = CSDFT_create_user_func("user_create_kernel", sizeof(struct csdft_user_data_defn), sizeof(FloatComplex)*n, 0, 1);

    /* Execute the function */
    status    = CSDFT_execute_user_func(user_func, md_init, md_kernel, md_null, CSDFT_BOARD_SEQUENCE | CSDFT_PROCESSOR_SEQUENCE) ;
    
    if (CSDFT_get_status() != CSDFT_NO_ERROR)
    {
    
       /* An error has occurred when processing kernel. Tell user the problem: */
       printf("%s", CSDFT_return_error_message(CSDFT_get_status()));
 
       /* Fail the test */
       passed = 0;
       return passed;
    
    }
    
    
    /* Describe the convolve */
    csdft_user_data.flags          = CSDFT_OPTIMAL_ORDER|CSDFT_COMPLEX|CSDFT_SINGLE;
    csdft_user_data.size_x         = n;
    csdft_user_data.scale          = 1.0/n;
    csdft_user_data.dimensions     = 1;
    csdft_user_data.number_of_ffts = num_ffts;
    
    
   /* Create the function that will perform the convolution */
   conv_func = CSDFT_create_user_func("user_convolve_perform_dft", sizeof(struct csdft_user_data_defn), sizeof(FloatComplex)*n*num_ffts, sizeof(FloatComplex)*n*num_ffts, 1);

   /* Execute the function */
   status = CSDFT_execute_user_func(conv_func, md_init, md_in, md_out, CSDFT_BOARD_SEQUENCE | CSDFT_PROCESSOR_SEQUENCE);


    /* Check that the input and output are almost the same */
    for (j = 0; j < num_ffts; j++)
    {
        if (!compareFft(&out[j*n], &in[j*n], 1, n))
        {
            printf("DFT %d failed for size %d\n", j+1, n);
            passed = 0;
            break;
        }
    }


    /* Free the memory */
    free(kernel);
    free(out);
    free(in);

    /* Free the memory descriptors */
    CSDFT_free(md_kernel);
    CSDFT_free(md_out);
    CSDFT_free(md_in);




    return passed;
}

/*
    Main
*/
int main(int argc, char *argv[])
{
    unsigned int s;
    unsigned int passed = 1;

    for (s = 128; s <= 8192; s = s*2)
    {
        if (!cs_dft_user_convolution(s))
        {
            printf("Failed 1D single precision complex user convolution of size %d\n\n", s);
            passed  = 0;
        }
    }

    printf("\n**** 1D complex single precision user convolution test: ");

    if (passed)
    {
        printf("passed\n");
    } else
    {
        printf("failed\n");
        return 1;
    }
    printf("\n");

    return 0;
}

