/*
 *
 *     Copyright (c) 2006, ClearSpeed Technology plc. All rights reserved.
 *
 * This material is provided "as is", with absolutely no warranty expressed or
 * implied. Any use is at your own risk. Permission to use or copy this software
 * for any purpose is hereby granted without fee, provided the above notices are
 * retained on all copies. Permission to modify the code and to distribute
 * modified code is granted, provided the above notices are retained, and a
 * notice that the code was modified is included with the above copyright notice.
 *
 */


/*!
  \file   cs_dft_demo_2d_r_fwd_rev.c
  \brief  Clearspeed DFT Library Test

  \date   30-January-2006

  This test demonstrates the creation of single precision 2d real plans
  and their execution. It cycles through sizes 128 to 1024 in powers of 2,
  creating random data and then allocating a memory descriptor to the data.
  It then creates 2 2d single precision plans, a forward real to complex and a
  backward complex to real.

  It executes one after the other, such that the resulting output should be
  almost identical to the original input.

*/

#include <math.h>
#include <stdlib.h>

#include "csdft.h"
#include "cs_complex.h"

/*
    Routines to help compare input and golden image
*/
unsigned int compare_value(float val, float gold_val)
{
    float diff;
    if ( fabsf(gold_val) < 0.001)
    {
        diff = fabsf(val-gold_val);
    } else
    {
        diff = fabsf((val-gold_val)/gold_val);
    }
    if (diff > 0.05)
    {
        return 0;
    } else
    {
        return 1;
    }
}


unsigned int compareFft(float *result, float *gold, unsigned int w, unsigned int h)
{
    unsigned int r,c;

    for (r = 0; r < h; r++)
        for (c = 0; c < w; c++)
        {
            if (!compare_value(result[r*w+c], gold[r*w+c]))
            {
                printf("Failed fft comparison locn, column %d, row %d\n", c, r);
                return 0;
            }
        }
    return 1;
}

/*
    Single precision complex 2d routine for size n
*/
unsigned int cs_dft_forward_reverse(unsigned int n)
{
    unsigned int i = 0, j = 0;
    unsigned int num_ffts = 3;
    unsigned int passed = 1;

    float v;

    float *in = NULL, *out = NULL, *inter = NULL;

    CSDftMemoryDescriptor md_in;
    CSDftMemoryDescriptor md_out;
    CSDftMemoryDescriptor md_inter;
    CSDftStatus           status;
    CSDftPlan             forward, reverse;


    /* malloc local space for ffts */
    in     = malloc(sizeof(float) * n * n * num_ffts);
    inter  = malloc(sizeof(FloatComplex) * (n/2+1) * n * num_ffts);
    out     = malloc(sizeof(float) * n * n * num_ffts);


    /* Generate some random input */
    srand(1);

    for (j = 0; j < num_ffts; j++)
    {
        for (i=0; i<n*n ; i++)
        {
            do {
                v  = (1.0*rand())/RAND_MAX;
            } while (v < 0.02);

            in[j*n*n+i] = v;
        }
    }


    /* Create the host memory descriptors that point to the pre-allocated in and out arrays */
    md_in  = CSDFT_system_to_memory_descriptor((void *)in,  (sizeof(float) * n * n)* num_ffts);
    md_out  = CSDFT_system_to_memory_descriptor((void *)out,  (sizeof(float) * n * n)* num_ffts);
    md_inter = CSDFT_system_to_memory_descriptor((void *)inter, (sizeof(FloatComplex) * (n/2+1) * n)* num_ffts);


    /*
     Create the plan that describes the DFTs, in this case an unscaled single
     precision complex forward transform and a corresponding scaled backward one
    */
    forward = CSDFT_create_plan_2d(CSDFT_FORWARD |CSDFT_NATURAL_ORDER|CSDFT_SINGLE|
                                   CSDFT_SRC_REAL | CSDFT_DST_COMPLEX, n, n, num_ffts, 1.0   ) ;
    reverse = CSDFT_create_plan_2d(CSDFT_BACKWARD|CSDFT_NATURAL_ORDER|CSDFT_DST_REAL| CSDFT_SRC_COMPLEX|
                                   CSDFT_SINGLE, n, n, num_ffts, 1.0/(n*n) ) ;

    /* Perform the DFTs by calling the synchronous interface */
    status = CSDFT_execute_dft(forward, md_in,  md_inter) ;
    status = CSDFT_execute_dft(reverse, md_inter, md_out) ;


    /* Check that the input and output are almost the same */
    for (j = 0; j < num_ffts; j++)
    {
        if (!compareFft(&out[j*n*n], &in[j*n*n], n, n))
        {
            printf("Dft %d failed for size %d\n", j+1, n);
            passed = 0;
            break;
        }
    }


    /* Free the memory */
    free(out);
    free(in);

    /* Free the memory descriptors */
    CSDFT_free(md_out);
    CSDFT_free(md_in);

    /* Free the plans */
    CSDFT_free_plan(forward);
    CSDFT_free_plan(reverse);


    return passed;
}

/*
    Main
*/
int main(int argc, char *argv[])
{
    unsigned int s;
    unsigned int passed = 1;

    for (s = 128; s <= 1024; s = s*2)
    {
        if (!cs_dft_forward_reverse(s))
        {
            printf("Failed 2D single precision real forward reverse dft of size %d x %d\n\n", s, s);
            passed  = 0;
        }
    }

    printf("\n**** 2D real single precision forward reverse test: ");

    if (passed)
    {
        printf("passed\n");
    } else
    {
        printf("failed\n");
        return 1;
    }
    printf("\n");

    return 0;
}

