/*
 *
 *     Copyright (c) 2006, ClearSpeed Technology plc. All rights reserved.
 *
 * This material is provided "as is", with absolutely no warranty expressed or
 * implied. Any use is at your own risk. Permission to use or copy this software
 * for any purpose is hereby granted without fee, provided the above notices are
 * retained on all copies. Permission to modify the code and to distribute
 * modified code is granted, provided the above notices are retained, and a
 * notice that the code was modified is included with the above copyright notice.
 *
 */

#ifndef _CSDFT_H
#define _CSDFT_H

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/*============================================================================*/
/* enums, struct and typedef declarations for CSDftMemoryDescriptors          */
/*============================================================================*/


enum _CSDftMemoryDescriptorType
{
    CSDFT_HOST      = 1,
    CSDFT_BOARD     = 2,
    CSDFT_PROCESSOR = 4
} ;

typedef enum _CSDftMemoryDescriptorType CSDftMemoryDescriptorType ;


struct CSDftMemoryDescriptorObj
{
    void *       data_pointer ;
    void *       data_pointer_imag ;
    unsigned int null_memory_descriptor ;
    unsigned int split_arrays ;
    unsigned int size_in_bytes ;
    unsigned int number_of_buffers ;

    CSDftMemoryDescriptorType location ;
    unsigned int board_instance ;
    unsigned int processor_instance ;
    unsigned int start_buffer ;
    unsigned int end_buffer ;

    unsigned int data_chip;         /* non-zero if this is a chip descriptor */
} ;

typedef struct CSDftMemoryDescriptorObj *CSDftMemoryDescriptor;

/*============================================================================*/
/* This allows us to have a predefined NULL Memory Descriptor to use as an    */
/* argument for user plans and still maintain a level of error checking for   */
/* other Memory Descriptor parameters to the CSDFT_execute_dft routine        */
/*============================================================================*/

extern struct CSDftMemoryDescriptorObj _CSDftNullMemoryDescriptor ;

#define CSDftNullMemoryDescriptor &_CSDftNullMemoryDescriptor

CSDftMemoryDescriptor CSDFT_get_null_descriptor(void) ;

/*============================================================================*/
/* enums and function prototypes for CSDft status routines                    */
/*============================================================================*/

enum _CSDftStatusTypes
{
    CSDFT_NO_ERROR,
    CSDFT_RUNNING,
    CSDFT_FINISHED,
    CSDFT_BOARD_ERROR,
    CSDFT_INVALID_BOARD,
    CSDFT_INVALID_PROCESSOR,
    CSDFT_MEMORY_ALLOCATION_ERROR,
    CSDFT_INVALID_MEMORY_DESCRIPTOR,
    CSDFT_INVALID_PLAN,
    CSDFT_INVALID_USER_FUNCTION,
    CSDFT_INVALID_SYMBOL,
    CSDFT_INVALID_SIZE,
    CSDFT_INVALID_FLAG,
    CSDFT_INVALID_PARAMETER
} ;

typedef enum _CSDftStatusTypes CSDftStatus ;


CSDftStatus CSDFT_get_status            ( void ) ;                  /* Returns last valid status */

const char *CSDFT_return_error_message  ( CSDftStatus status ) ;    /* Returns a NULL terminated string that textualy describes a given error code */


/*============================================================================*/
/* Function prototypes for CSDftMemoryDescriptor creation routines            */
/*============================================================================*/
CSDftMemoryDescriptor CSDFT_malloc_host ( unsigned int size_in_bytes) ;

/*
CSDftMemoryDescriptor CSDFT_malloc      ( unsigned int size_in_bytes) ;

CSDftMemoryDescriptor CSDFT_malloc_processor
                                        ( unsigned int size_in_bytes,
                                          unsigned int board_instance,
                                          unsigned int processor_instance ) ;
*/
CSDftStatus           CSDFT_free        ( CSDftMemoryDescriptor handle ) ;


/*============================================================================*/
/* Function prototypes for CSDftMemoryDescriptor manipulation routines        */
/*============================================================================*/

void *CSDFT_memory_descriptor_to_system( CSDftMemoryDescriptor handle ) ;

CSDftMemoryDescriptor CSDFT_system_to_memory_descriptor
                                       ( void *       source_data,
                                         unsigned int size_in_bytes) ;

CSDftMemoryDescriptor CSDFT_processor_to_memory_descriptor
                                       ( unsigned int source_data,
                                         unsigned int size_in_bytes,
                                         unsigned int board_instance,
                                         unsigned int processor_instance ) ;

CSDftMemoryDescriptor CSDFT_system_split_array_to_memory_descriptor
                                       ( void *       source_data_real,
                                         void *       source_data_imag,
                                         unsigned int size_in_bytes ) ;

CSDftStatus CSDFT_memory_descriptor_copy
                                       ( const CSDftMemoryDescriptor source,
                                               CSDftMemoryDescriptor destination,
                                         const unsigned int          size_in_bytes ) ;


/*============================================================================*/
/* defines, enums, struct and typedef declarations for CSDft Processor ID     */
/*============================================================================*/

#define CSDFT_PROCESSOR_ID_BOARD_SHIFT              0x0          /* No shift  */
#define CSDFT_PROCESSOR_ID_BOARD_SEQUENCE_SHIFT     0x7          /* bit 7     */
#define CSDFT_PROCESSOR_ID_BOARD_MASK               0x7f         /* bits 0..6 */
#define CSDFT_PROCESSOR_ID_BOARD_SEQUENCE_MASK      0x80         /* bit 7     */

#define CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT          0x8          /* 8 bit shift, represents bits 8..E */
#define CSDFT_PROCESSOR_ID_PROCESSOR_SEQUENCE_SHIFT 0xf          /* 15 bit shift, represents bit 15   */
#define CSDFT_PROCESSOR_ID_PROCESSOR_MASK           0x7f00       /* bits 8..15 */
#define CSDFT_PROCESSOR_ID_PROCESSOR_SEQUENCE_MASK  0x8000       /* bit 15     */

#define CSDFT_PROCESSOR_ID_SHIFT(value, shift)       (value << shift)
#define CSDFT_PROCESSOR_ID_MASK(value, mask, shift) ((value & mask) >> shift)

enum _CSDftUserProcessor
{
    /* _SEQUENCE indicates to the library that it should use the next available board */
    CSDFT_BOARD_0        = CSDFT_PROCESSOR_ID_SHIFT(0x0, CSDFT_PROCESSOR_ID_BOARD_SHIFT),                   /* bit 0 */
    CSDFT_BOARD_1        = CSDFT_PROCESSOR_ID_SHIFT(0x1, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_2        = CSDFT_PROCESSOR_ID_SHIFT(0x2, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_3        = CSDFT_PROCESSOR_ID_SHIFT(0x3, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_4        = CSDFT_PROCESSOR_ID_SHIFT(0x4, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_5        = CSDFT_PROCESSOR_ID_SHIFT(0x5, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_6        = CSDFT_PROCESSOR_ID_SHIFT(0x6, CSDFT_PROCESSOR_ID_BOARD_SHIFT),
    CSDFT_BOARD_7        = CSDFT_PROCESSOR_ID_SHIFT(0x7, CSDFT_PROCESSOR_ID_BOARD_SHIFT),                   /* bit 6 */
    CSDFT_BOARD_SEQUENCE = CSDFT_PROCESSOR_ID_SHIFT(0x1, CSDFT_PROCESSOR_ID_BOARD_SEQUENCE_SHIFT),          /* bit 7 */

    /* _SEQUENCE indicates to the library that it should use the next available processor */
    CSDFT_PROCESSOR_0        = CSDFT_PROCESSOR_ID_SHIFT(0x0, CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT),           /* bit 8  */
    CSDFT_PROCESSOR_1        = CSDFT_PROCESSOR_ID_SHIFT(0x1, CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT),
    CSDFT_PROCESSOR_2        = CSDFT_PROCESSOR_ID_SHIFT(0x2, CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT),
    CSDFT_PROCESSOR_3        = CSDFT_PROCESSOR_ID_SHIFT(0x3, CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT),           /* bit 11 */
    CSDFT_PROCESSOR_SEQUENCE = CSDFT_PROCESSOR_ID_SHIFT(0x1, CSDFT_PROCESSOR_ID_PROCESSOR_SEQUENCE_SHIFT)   /* bit 15 */
} ;

typedef enum _CSDftUserProcessor CSDftUserProcessor;

#define  CSDFT_create_user_processor(board, processor) \
 ((board<<CSDFT_PROCESSOR_ID_BOARD_SHIFT)| (processor<<CSDFT_PROCESSOR_ID_PROCESSOR_SHIFT))

/*============================================================================*/
/* defines, enums, struct and typedef declarations for CSDftPlan              */
/*============================================================================*/

#define CSDFT_DIRECTION                 0                           /* 1 bit    */
#define CSDFT_ORDER                     1                           /* 1 bit    */
#define CSDFT_SRC_PRECISION             2                           /* 2 bits   */
#define CSDFT_SRC_TYPE                  4                           /* 1 bits   */
#define CSDFT_DST_PRECISION             5                           /* 2 bits   */
#define CSDFT_DST_TYPE                  7                           /* 1 bit    */

#define CSDFT_SRC_FORMAT                30                          /* 1 bit    */
#define CSDFT_DST_FORMAT                31                          /* 1 bit    */

#define CSDFT_DIRECTION_MASK            0x1                         /* bit 0    */
#define CSDFT_ORDER_MASK                0x2                         /* bit 1    */
#define CSDFT_SRC_PRECISION_MASK        0xC                         /* bit 2&3  */
#define CSDFT_SRC_TYPE_MASK             0x10                        /* bit 4    */
#define CSDFT_DST_PRECISION_MASK        0x60                        /* bit 5&6  */
#define CSDFT_DST_TYPE_MASK             0x80                        /* bit 7    */

#define CSDFT_CONVOLVE_TRANSFORM_MASK   0x20000000                  /* bit 29   */
#define CSDFT_SRC_FORMAT_MASK           0x40000000                  /* bit 30   */
#define CSDFT_DST_FORMAT_MASK           0x80000000                  /* bit 31   */

#define CSDFT_FORWARD                   (0 << CSDFT_DIRECTION)      /* Default is FORWARD (0)   */
#define CSDFT_BACKWARD                  (1 << CSDFT_DIRECTION)      /*                    (1)   */

#define CSDFT_NATURAL_ORDER             (0 << CSDFT_ORDER)          /* Defaul is NATURAL        */
#define CSDFT_OPTIMAL_ORDER             (1 << CSDFT_ORDER)
/*#define CSDFT_BIT_REVERSED_ORDER      (2 << CSDFT_ORDER)*/

#define CSDFT_SRC_BYTE                  (0 << CSDFT_SRC_PRECISION)  /* Default is BYTE          */
#define CSDFT_SRC_SHORT                 (1 << CSDFT_SRC_PRECISION)
#define CSDFT_SRC_SINGLE                (2 << CSDFT_SRC_PRECISION)
#define CSDFT_SRC_DOUBLE                (3 << CSDFT_SRC_PRECISION)

#define CSDFT_SRC_REAL                  (0 << CSDFT_SRC_TYPE)       /* Default is REAL          */
#define CSDFT_SRC_COMPLEX               (1 << CSDFT_SRC_TYPE)

#define CSDFT_DST_BYTE                  (0 << CSDFT_DST_PRECISION)  /* Default is BYTE          */
#define CSDFT_DST_SHORT                 (1 << CSDFT_DST_PRECISION)
#define CSDFT_DST_SINGLE                (2 << CSDFT_DST_PRECISION)
#define CSDFT_DST_DOUBLE                (3 << CSDFT_DST_PRECISION)

#define CSDFT_DST_REAL                  (0 << CSDFT_DST_TYPE)       /* Default is REAL          */
#define CSDFT_DST_COMPLEX               (1 << CSDFT_DST_TYPE)

#define CSDFT_SRC_INTERLEAVED           (0 << CSDFT_SRC_FORMAT)     /* Default is INTERLEAVED   */
#define CSDFT_SRC_SPLIT_ARRAY           (1 << CSDFT_SRC_FORMAT)
#define CSDFT_DST_INTERLEAVED           (0 << CSDFT_DST_FORMAT)     /* Default is INTERLEAVED   */
#define CSDFT_DST_SPLIT_ARRAY           (1 << CSDFT_DST_FORMAT)

/* Combined source and destination */
#define CSDFT_BYTE                      (CSDFT_SRC_BYTE|CSDFT_DST_BYTE)
#define CSDFT_SHORT                     (CSDFT_SRC_SHORT|CSDFT_DST_SHORT)
#define CSDFT_SINGLE                    (CSDFT_SRC_SINGLE|CSDFT_DST_SINGLE)
#define CSDFT_DOUBLE                    (CSDFT_SRC_DOUBLE|CSDFT_DST_DOUBLE)
#define CSDFT_REAL                      (CSDFT_SRC_REAL|CSDFT_DST_REAL)
#define CSDFT_COMPLEX                   (CSDFT_SRC_COMPLEX|CSDFT_DST_COMPLEX)
#define CSDFT_INTERLEAVED               (CSDFT_SRC_INTERLEAVED|CSDFT_DST_INTERLEAVED)
#define CSDFT_SPLIT_ARRAY               (CSDFT_SRC_SPLIT_ARRAY|CSDFT_DST_SPLIT_ARRAY)
#define CSDFT_TRANSFORM_KERNEL           CSDFT_CONVOLVE_TRANSFORM_MASK


typedef unsigned int CSDftFlags;


struct CSDftPlanObj
{
    unsigned int flags ;
    unsigned int dimension ;
    unsigned int direction ;
    unsigned int order ;
    unsigned int src_precision ;
    unsigned int src_type ;
    unsigned int src_format ;
    unsigned int src_sample_size ;
    unsigned int dst_precision ;
    unsigned int dst_type ;
    unsigned int dst_format ;
    unsigned int dst_sample_size ;
    unsigned int size[3] ;
    unsigned int number_of_ffts ;
    double       scale ;
/*    char *       user_exec_routine ; */
    unsigned int *user_control_func ;
} ;

typedef struct CSDftPlanObj * CSDftPlan ;

struct CSDftUserFuncObj
{
    unsigned int common_buf_size;
    unsigned int src_buf_size;
    unsigned int dst_buf_size;
    unsigned int num_blocks;
    unsigned int *user_control_func ;
} ;

typedef struct CSDftUserFuncObj * CSDftUserFunc;

typedef unsigned int CSDftExecDescriptor ;


/*============================================================================*/
/* Function prototypes for CSDftPlan creation and execution routines          */
/*============================================================================*/

CSDftPlan CSDFT_create_plan_1d      ( CSDftFlags   flags,
                                      unsigned int size,
                                      unsigned int number_of_ffts,
                                      double       scale ) ;

CSDftPlan CSDFT_create_plan_2d      ( CSDftFlags   flags,
                                      unsigned int x_size,
                                      unsigned int y_size,
                                      unsigned int number_of_ffts,
                                      double       scale ) ;

CSDftPlan CSDFT_create_plan_3d      ( CSDftFlags   flags,
                                      unsigned int x_size,
                                      unsigned int y_size,
                                      unsigned int z_size,
                                      unsigned int number_of_ffts,
                                      double       scale ) ;

CSDftStatus CSDFT_free_plan         ( CSDftPlan handle ) ;

CSDftStatus CSDFT_free_user_func    ( CSDftUserFunc handle ) ;

CSDftStatus CSDFT_execute_dft       ( const CSDftPlan       plan,
                                      CSDftMemoryDescriptor source,
                                      CSDftMemoryDescriptor destination ) ;
/*
CSDftStatus CSDFT_execute_dft_async ( const CSDftPlan           plan,
                                      const CSDftExecDescriptor exec,
                                      CSDftMemoryDescriptor     source,
                                      CSDftMemoryDescriptor     destination ) ;

CSDftStatus CSDFT_wait              ( CSDftExecDescriptor exec ) ;

CSDftStatus CSDFT_completed         ( CSDftExecDescriptor exec ) ;
*/

CSDftPlan CSDFT_create_convolution_plan_1d ( CSDftFlags   flags,
                                             unsigned int size,
                                             unsigned int number_of_ffts,
                                             double       scale ) ;

CSDftPlan CSDFT_create_convolution_plan_2d ( CSDftFlags   flags,
                                             unsigned int x_size,
                                             unsigned int y_size,
                                             unsigned int number_of_ffts,
                                             double       scale ) ;
/*
CSDftPlan CSDFT_create_convolution_plan_3d ( CSDftFlags   flags,
                                             unsigned int x_size,
                                             unsigned int y_size,
                                             unsigned int z_size,
                                             unsigned int number_of_ffts,
                                             double       scale ) ;
*/
CSDftStatus CSDFT_execute_convolution ( const CSDftPlan       plan,
                                        CSDftMemoryDescriptor kernel,
                                        CSDftMemoryDescriptor source,
                                        CSDftMemoryDescriptor destination ) ;

CSDftUserFunc CSDFT_create_user_func  ( const char * user_exec_routine,
                                        unsigned int common_size,
                                        unsigned int src_size,
                                        unsigned int dst_size,
                                        unsigned int num_blocks ) ;

CSDftStatus CSDFT_execute_user_func   ( const CSDftUserFunc   func,
                                        CSDftMemoryDescriptor common,
                                        CSDftMemoryDescriptor source,
                                        CSDftMemoryDescriptor destination,
                                        CSDftUserProcessor    processor ) ;

#ifdef __cplusplus
}  /* extern "C" */
#endif /* __cplusplus */

#endif /* CSDFT_H */

